"""
The :py:mod:`vc2_conformance.py2x_compat` module provides backported
implementations of various functions from Python 3 which are used by this
software.


.. py:function:: zip_longest

    In Python 3.x an alias for :py:func:`itertools.zip_longest`, in Python
    2.x, an alias for ``itertools.izip_longest``.


.. py:function:: get_terminal_size

    In Python 3.x an alias for :py:func:`shutil.get_terminal_size`, in Python
    2.x, a dummy function always returning ``(80, 20)``.


.. py:function:: wraps

    In Python 3.x an alias for :py:func:`functools.wraps`. In Python 2.x, an
    alternative implementation of ``functools.wraps`` which includes the Python
    3.x behaviour of setting the ``__wrapped__`` attribute to allow
    introspection of wrapped functions (see :py:func:`unwrap`).


.. py:function:: unwrap

    In Python 3.x an alias for :py:func:`inspect.unwrap`. In Python 2.x a
    backported implementation of that function. Relies on the backported
    :py:func:`wraps` implementation provided by this module.


.. py:function:: quote

    In Python 3.x an alias for :py:func:`shlex.quote`, in Python
    2.x, an alias for ``pipes.quote``.


.. py:data:: string_types

    A tuple enumerating the native string-like types.  In Python 3.x, ``(str,
    )``, in Python 2.x, ``(str, unicode)``.


.. py:function:: gcd

    In Python 3.x an alias for :py:func:`math.gcd`, in Python
    2.x, an alias for ``fractions.gcd``.


.. py:function:: zip

    In Python 3.x an alias for :py:func:`zip`, in Python
    2.x, an alias for ``itertools.izip``.


.. py:function:: makedirs

    In Python 3.x an alias for :py:func:`os.makedirs`. In Python 2.x, a
    backport of this function which includes the ``exist_ok`` argument.


.. py:function:: FileType

    In Python 3.x an alias for :py:class:`argparse.FileType`. In Python 2.x, a
    wrapper around :py:class:`argparse.FileType` adding support for the
    'encoding' keyword argument when opening with mode "r".

"""

__all__ = [
    "zip_longest",
    "get_terminal_size",
    "wraps",
    "unwrap",
    "quote",
    "string_types",
    "gcd",
    "zip",
    "makedirs",
    "FileType",
]

import os

import sys

import io


try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest


try:
    from shutil import get_terminal_size
except ImportError:
    # Crude fallback implementation (fixed assumed size)
    def get_terminal_size(fallback=(80, 20)):
        return fallback


if sys.version_info[0] >= 3:
    from functools import wraps
else:
    # In Python 2.x the 'functoos.wraps' function does not set the __wrapped__
    # attribute of wrapped functions. This means introspection tools are unable
    # to find the originally wrapped function.
    from functools import wraps as _wraps

    def wraps(wrapped, *args, **kwargs):
        def wrapper(f):
            f = _wraps(wrapped, *args, **kwargs)(f)
            f.__wrapped__ = wrapped
            return f

        return wrapper

    wraps = wraps(_wraps)(wraps)


try:
    from inspect import unwrap
except ImportError:

    def unwrap(func, stop=None):
        if stop is None:

            def stop(f):
                return False

        while hasattr(func, "__wrapped__") and not stop(func):
            func = func.__wrapped__

        return func


try:
    from shlex import quote
except ImportError:
    from pipes import quote


try:
    string_types = (str, unicode)
except NameError:
    string_types = (str,)

try:
    from math import gcd  # Python >= 3.5
except ImportError:
    from fractions import gcd  # Python < 3.5

try:
    from itertools import izip as zip  # Python 2.x
except ImportError:
    zip = zip

try:
    os.makedirs(os.path.dirname(__file__), exist_ok=True)
    makedirs = os.makedirs
except TypeError:

    def makedirs(name, mode=0o777, exist_ok=False):
        if exist_ok and os.path.isdir(name):
            return
        else:
            os.makedirs(name, mode)


if sys.version_info[0] >= 3:
    from argparse import FileType
else:
    # In Python 2.x, the FileType class does not accept an 'encoding' argument.
    # We reimplement this feature here.
    from argparse import FileType as _FileType

    class FileType(_FileType):
        def __init__(self, mode, *args, **kwargs):
            if mode == "r" and "encoding" in kwargs:
                self._encoding = kwargs.pop("encoding")
            else:
                self._encoding = None
            super(FileType, self).__init__(mode, *args, **kwargs)

        def __call__(self, filename):
            file_obj = super(FileType, self).__call__(filename)
            if self._encoding is None:
                return file_obj
            else:
                # Special case backported for Python 2.x
                return io.open(filename, "r", encoding=self._encoding)



"""
The exceptions defined in :py:mod:`vc2_conformance.encoder.exceptions` derive
from :py:exc:`UnsatisfiableCodecFeaturesError` and are thrown when the
presented encoder configuration makes encoding impossible. These exceptions
provide detailed explanations of why encoding was not possible.

.. autoexception:: UnsatisfiableCodecFeaturesError
    :members:

"""

from vc2_conformance.string_utils import wrap_paragraphs


class UnsatisfiableCodecFeaturesError(ValueError):
    """
    Base class for exceptions thrown by the encoder when it is unable to
    generate a stream in the desired format due to some invalid
    :py:class:`~vc2_conformance.codec_features.CodecFeatures` configuration.
    """

    def __str__(self):
        return wrap_paragraphs(self.explain()).partition("\n")[0]

    def explain(self):
        """
        Produce a detailed human readable explanation of the failure.

        Should return a string which can be re-linewrapped by
        :py:func:`vc2_conformance.string_utils.wrap_paragraphs`.

        The first line will be used as a summary when the exception is printed
        using :py:func:`str`.
        """
        raise NotImplementedError()


class MissingQuantizationMatrixError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when the codec features specified did not include a quantization
    matrix when one was required.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            No default quantization matrix is available for the wavelet
            transform specified for {} and no custom quantization matrix was
            provided.

            * wavelet_index: {} ({:d})
            * wavelet_index_ho: {} ({:d})
            * dwt_depth: {}
            * dwt_depth_ho: {}
        """.format(
            codec_features["name"],
            codec_features["wavelet_index"].name,
            codec_features["wavelet_index"],
            codec_features["wavelet_index_ho"].name,
            codec_features["wavelet_index_ho"],
            codec_features["dwt_depth"],
            codec_features["dwt_depth_ho"],
        )


class PictureBytesSpecifiedForLosslessModeError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when the 'picture_bytes' field of a
    :py:class:`vc2_conformance.codec_features.CodecFeatures` is set at the same
    time as the lossless field being true.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            The codec configuration {} specifies a lossless format but did not
            omit picture_bytes (it is set to {}).
        """.format(
            codec_features["name"],
            codec_features["picture_bytes"],
        )


class InsufficientHQPictureBytesError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when the 'picture_bytes' field of a
    :py:class:`vc2_conformance.codec_features.CodecFeatures` is set too low for
    the coding options chosen.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            The codec configuration {} specifies picture_bytes as {} but this
            is too small.

            A {} by {} slice high quality profile encoding must allow at least
            4*{}*{} = {} bytes. (That is, 1 byte for the qindex field and 3
            bytes for the length fields of each high quality slice (13.5.4))
        """.format(
            codec_features["name"],
            codec_features["picture_bytes"],
            codec_features["slices_x"],
            codec_features["slices_y"],
            codec_features["slices_x"],
            codec_features["slices_y"],
            4 * codec_features["slices_x"] * codec_features["slices_y"],
        )


class InsufficientLDPictureBytesError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when the 'picture_bytes' field of a
    :py:class:`vc2_conformance.codec_features.CodecFeatures` is set too low for
    the coding options chosen.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            The codec configuration {} specifies picture_bytes as {} but this
            is too small.

            A {} by {} slice low delay profile encoding must allow at least
            {}*{} = {} bytes. (That is, 7 bits for the qindex field and 1 bit
            for the slice_y_length field of each low delay slice (13.5.3.1))
        """.format(
            codec_features["name"],
            codec_features["picture_bytes"],
            codec_features["slices_x"],
            codec_features["slices_y"],
            codec_features["slices_x"],
            codec_features["slices_y"],
            codec_features["slices_x"] * codec_features["slices_y"],
        )


class LosslessUnsupportedByLowDelayError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when lossless coding is chosen for the low delay profile (which only
    supports lossy coding).
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            Low delay profile does not support lossless encoding for {}.
        """.format(
            codec_features["name"],
        )


class IncompatibleLevelAndVideoFormatError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when the codec features specified a particular VC-2 level which is
    incompatible with the video format specified.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            Level {} ({:d}) specified for {} but the video format requested is
            not permitted by this level.
        """.format(
            codec_features["level"].name,
            codec_features["level"],
            codec_features["name"],
        )


class IncompatibleLevelAndExtendedTransformParametersError(
    UnsatisfiableCodecFeaturesError
):
    """
    Thrown when the codec features specified a particular VC-2 level which is
    incompatible with the extended transform parameter encoding required.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            Level {} ({:d}) specified for {} but the asymmetric transform type
            specified is not not permitted by this level.

            * wavelet_index: {} ({:d})
            * wavelet_index_ho: {} ({:d})
            * dwt_depth: {}
            * dwt_depth_ho: {}
        """.format(
            codec_features["level"].name,
            codec_features["level"],
            codec_features["name"],
            codec_features["wavelet_index"].name,
            codec_features["wavelet_index"],
            codec_features["wavelet_index_ho"].name,
            codec_features["wavelet_index_ho"],
            codec_features["dwt_depth"],
            codec_features["dwt_depth_ho"],
        )


class IncompatibleLevelAndDataUnitError(UnsatisfiableCodecFeaturesError):
    """
    Thrown when a sequence of data unit types is requested which is not allowed
    by the current level.
    """

    def explain(self):
        (codec_features,) = self.args

        return """
            Internal error in conformance software (sorry about that!).

            A test case was generated for {} whose level, {} ({:d}), prohibited
            the sequence of data units the test case required.
        """.format(
            codec_features["name"],
            codec_features["level"].name,
            codec_features["level"],
        )



r"""
The :py:mod:`vc2_conformance.bitstream.serdes` module provides a framework for
transforming a set of functions designed to process a bitstream (e.g. the VC-2
specification's pseudocode) into general-purpose bitstream serialisers,
deserialisers and analysers.

The following sections introduce the design and operation of this module in
detail.

A basic bitstream serialiser
````````````````````````````

The VC-2 specification describes the bitstream and decoding process in a series
of pseudocode functions such as the following::

    frame_size(video_parameters):
        custom_dimensions_flag = read_bool()
        if(custom_dimensions_flag == True)
            video_parameters[frame_width] = read_uint()
            video_parameters[frame_height] = read_uint()

To see how this definition might be transformed into a general purpose
bitstream serialiser we must transform this definition of a program which
*reads* a VC-2 bitstream into one which *writes* one.

We start by replacing all of the ``read_*`` functions with equivalent
``write_*`` functions (which we define here as returning the value that they
write)::

    frame_size(video_parameters):
        custom_dimensions_flag = write_bool(???)
        if(custom_dimensions_flag == True)
            video_parameters[frame_width] = write_uint(???)
            video_parameters[frame_height] = write_uint(???)

Next we need to define what values we'd like writing by replacing the ``???``
placeholders with a suitable global variables like so::

    new_custom_dimensions_flag = True
    new_frame_width = 1920
    new_frame_height = 1080

    frame_size(video_parameters):
        custom_dimensions_flag = write_bool(new_custom_dimensions_flag)
        if(custom_dimensions_flag == True)
            video_parameters[frame_width] = write_uint(new_frame_width)
            video_parameters[frame_height] = write_uint(new_frame_height)

We have now transformed the VC-2 pseudocode bitstream *reader* function into a
*writer* function. What's more, by just changing the values of global variables
we created it is possible to use this function as a general-purpose bitstream
serialiser.


A basic deserialiser
````````````````````

Unfortunately, the original bitstream reader pseudocode from the VC-2
specification is not quite usable as a general-purpose bitstream deserialiser:

* The reader does not capture every value read from the bitstream in a variable
  we can later examine (e.g. the ``custom_dimensions_flag`` is kept in a local
  variable and not returned).
* The values which are captured are stored in a structure designed to aid
  decoding and not necessarily to faithfully describe a bitstream.

Lets create a new version of the reader function which overcomes these
limitations. We redefine the ``read_*`` functions to take an additional
argument naming a global variable where the read values will be stored, in
addition to being returned, giving the following pseudocode::

    read_custom_dimensions_flag = None
    read_frame_width = None
    read_frame_height = None

    frame_size(video_parameters):
        custom_dimensions_flag = read_bool(read_custom_dimensions_flag)
        if(custom_dimensions_flag == True)
            video_parameters[frame_width] = read_uint(read_frame_width)
            video_parameters[frame_height] = read_uint(read_frame_height)

This small change ensures that every value read from the bitstream is captured
in a global variable which we can later examine and which is orthogonal to
whatever data structures the VC-2 pseudocode might otherwise use.


An introduction to the :py:class:`SerDes` interface
---------------------------------------------------

The similarities between the transformations required to turn the VC-2
pseudocode into general purpose serialisers and deserialisers should be fairly
clear. In fact, the only difference between the two is that in one the
functions are called ``read_*`` and in the other they're called ``write_*``. In
both cases, the ``read_*`` and ``write_*`` functions take essentially the same
arguments: a name of a global variable.

This module defines the :py:class:`SerDes` interface which can be used by the
VC-2 pseudocode specifications to drive both bitstream serialisation and
deserialisation. To use it, we replace the ``read_*`` or ``write_*`` calls
with ``serdes.*`` calls.

Translating the ``frame_size`` function into valid Python and taking a
:py:class:`SerDes` instance as an argument we arrive at the following code::

    def frame_size(serdes, video_parameters):
        custom_dimensions_flag = serdes.bool("custom_dimensions_flag")
        if(custom_dimensions_flag == True)
            video_parameters["frame_width"] = serdes.uint("frame_width")
            video_parameters["frame_height"] = serdes.uint("frame_height")

To deserialise (read) a bitstream we use the :py:class:`Deserialiser`
implementation of :py:class:`SerDes` like so::

    >>> from vc2_conformance.bitstream import BitstreamReader, Deserialiser
    >>> reader = BitstreamReader(open("frame_size_snippet.bin", "rb"))

    >>> with Deserialiser(reader) as des:
    ...     frame_size(des, {})
    >>> des.context
    {"custom_dimensions_flag": True, "frame_width": 1920, "frame_height": 1080}

The :py:attr:`SerDes.context` property is a :py:class:`dict` which contains
each of the values read from the bitstream (named as per the calls to the
various :py:class:`SerDes` methods).

In the nomenclature of this module, this *context* dictionary holds values for
each of the *target names* specified by :py:meth:`SerDes.bool`,
:py:meth:`SerDes.uint` etc.

Values to be serialised should be structured into a context dictionary of
similar shape and passed to a :py:class:`Serialiser`::

    >>> from vc2_conformance.bitstream import BitstreamWriter, Serialiser
    >>> writer = BitstreamWriter(open("frame_size_snippet.bin", "wb"))

    >>> context = {"custom_dimensions_flag": True, "frame_width": 1920, "frame_height": 1080}
    >>> with Serialiser(writer, context) as ser:
    ...     frame_size(ser, {})

In this example a bitstream containing a '1' followed by the variable-length
integers '1920' and '1080' would be written to the bitstream.


Verification
````````````

The :py:class:`SerDes` implementations perform various 'sanity checks' during
serialisation and deserialisation to ensure that the values passed in or
returned have a 1:1 correspondence with values in the bitstream.

* When values are read during deserialisation, :py:class:`Deserialiser` checks
  that names are not re-used, guaranteeing that if a value appears in the
  bitstream it also appears in the output context dictionary (and are not
  later overwritten).
* When values are written during serialisation, :py:class:`Serialiser` checks
  that every value in the context dictionary is used exactly once, ensuring
  that every value provided is represented in the bitstream.
* During serialisation, values are also checked to ensure they can be
  correctly represented by the bitstream encoding. For example, providing a
  negative value to :py:meth:`~SerDes.uint` will fail.


Representing hierarchy
``````````````````````

The VC-2 bitstream does not represent a flat collection of values but rather a
hierarchy. The :py:class:`SerDes` interface provides additional facilities to
allow this structure to be recreated in the deserialised representation, making
it easier to inspect and describes bitstreams in their deserialised form.

For example, in the VC-2 specification, the ``source_parameters`` function
(11.4) is defined by a series of functions which each read the values
relating to a particular video feature such as the ``frame_size`` function
we've seen above. To collect together related values we can use
:py:meth:`SerDes.subcontext` to create nested context dictionaries::

    def source_parameters(serdes):
        video_parameters = {}
        with serdes.subcontext("frame_size"):
            frame_size(serdes, video_parameters)
        with serdes.subcontext("color_diff_sampling_format"):
            color_diff_sampling_format(serdes, video_parameters)
        # ...
        return video_parameters

This results in a nested dictionary structure::

    >>> with Deserialiser(reader) as des:
    ...     video_parameters = source_parameters(des)

    >>> from pprint import pprint
    >>> pprint(des.context)
    {
        "frame_size": {
            "custom_dimensions_flag": True,
            "frame_width": 1920,
            "frame_height": 1080,
        },
        "color_diff_sampling_format": {
            "custom_color_diff_format_flag": False,
        },
        # ...
    }

When used with :py:mod:`vc2_conformance.fixeddict`, :py:class:`SerDes` also
makes it possible to define custom dictionary types for each part of the
hierarchy using the :py:func:`context_type` decorator. Benefits include:

* Improved 'pretty-printed' string representations.
* Additional checks that unexpected values are not used accidentally in the bitstream.

For example, here's how the ``parse_info`` header (10.5.1) might be represented::

    from vc2_conformance.fixeddict import fixeddict, Entry
    from vc2_conformance.formatters import Hex
    from vc2_data_tables import ParseCodes, PARSE_INFO_PREFIX
    ParseInfo = fixeddict(
        "ParseInfo",
        Entry("parse_info_prefix", formatter=Hex(8)),
        Entry("parse_code", enum=ParseCodes),
        Entry("next_parse_offset"),
        Entry("previous_parse_offset"),
    )

    @context_type(ParseInfo)
    def parse_info(serdes, state):
        serdes.nbits(4*8, "parse_info_prefix")
        state["parse_code"] = serdes.nbits(8, "parse_code")
        state["next_parse_offset"] = serdes.nbits(32, "next_parse_offset")
        state["previous_parse_offset"] = serdes.nbits(32, "previous_parse_offset")

Using the above we can quickly create structures ready for serialisation::

    >>> context = ParseInfo(
    ...     parse_info_prefix=PARSE_INFO_PREFIX,
    ...     parse_code=ParseCodes.end_of_sequence,
    ...     next_parse_offset=0,
    ...     previous_parse_offset=1234,
    ... )
    >>> with Deserialiser(writer, context) as des:
    ...     parse_info(des, {})

We also benefit from improved string formatting when deserialising values::

    >>> with Deserialiser(reader) as des:
    ...     parse_info(des, {})
    >>> str(des.context)
    ParseInfo:
      parse_info_prefix: 0x42424344
      parse_code: end_of_sequence (0x10)
      next_parse_offset: 0
      previous_parse_offset: 1234

Representing arrays
```````````````````

The VC-2 bitstream format includes a number of array-like fields, for example
arrays of transform coefficients within slices. Rather than defining unique
names for every array value, :py:class:`SerDes` allows values to be declared as
lists. For example::

    def list_example(serdes):
        serdes.declare_list("three_values")
        serdes.uint("three_values")
        serdes.uint("three_values")
        serdes.uint("three_values")

When deserialising, the result will look like::

    >>> with Deserialiser(reader) as des:
    ...     list_example(des)
    >>> des.context
    {"three_values": [100, 200, 300]}

Likewise, when serialising, a list of values (of the correct length) should
also be provided:

    >>> context = {"three_values": [10, 20, 30]}
    >>> with Deserialiser(writer, context) as des:
    ...     list_example(des)

As usual, the :py:class:`SerDes` classes will verify that the correct number of
values is present and will throw exceptions when too many or too few are
provided.


Computed values
```````````````

In some circumstances, when interpreting a deserialised bitstream it may be
necessary to know information computed by an earlier part of the bitstream. For
example, the dimensions of a slice depend on numerous video formatting options.
To avoid error-prone reimplementation of these calculations it is possible to
use :py:meth:`SerDes.computed_value` to add values to the context dictionary
which do not appear in the bitstream. For example::

    def ld_slice(serdes, state, sx, sy):
        serdes.computed_value("_slices_x", state["slices_x"])
        serdes.computed_value("_slices_y", state["slices_y"])
        # ...

The computed value will be set in the context dictionary regardless of whether
serialisation or deserialisation is taking place and any existing value is
always ignored.

.. note::

    It is recommended that by convention computed value target names are
    prefixed or suffixed with an underscore.


Default values during serialisation
```````````````````````````````````

As discussed above, the default behaviour of the :py:class:`Serialiser` is to
require that every value in the bitstream is provided in the context dictionary
to make it explicit what is being serialised. In certain cases, however, it
may be desirable for certain values to be filled in automatically. For
example:

* For pre-filling constants like the parse_info prefix.
* For use in unit tests where only certain bitstream fields' values are of
  importance (and assigning defaults for the remainder makes the code clearer).
* For providing default (e.g. zero) values for padding fields

To facilitate this, the :py:class:`Serialiser` class may be passed a default
value lookup like so::

    >>> default_values = {
    ...     ParseInfo: {
    ...         "parse_info_prefix": PARSE_INFO_PREFIX,
    ...         "parse_code": ParseCodes.end_of_sequence,
    ...         "next_parse_offset": 0,
    ...         "previous_parse_offset": 0,
    ...     },
    ... }

    >>> writer = BitstreamWriter(open("frame_size_snippet.bin", "wb"))
    >>> context = ParseInfo(
    ...     parse_code=ParseCodes.end_of_sequence,
    ...     previous_parse_offset=123,
    ... )
    >>> with Serialiser(writer, context, default_values=default_values) as ser:
    ...     parse_info(ser, {})

The ``default_values`` lookup should provide a separate set of default values
for each context dictionary type. See
:py:data:`vc2_conformance.bitstream.vc2_fixeddicts.fixeddict_default_values`
for a complete example.

For arrays/lists of values, the default value provided will be usd to populate
array elements and not to provide a default for the list as a whole.

Where a default value is not found in the lookup, a :py;exc:`KeyError` will be
thrown as usual. This behaviour allows a partial set of default values to be
provided (e.g. providing defaults only for padding values) while still
validating that the provided input is correct.


API
---

.. autoclass:: SerDes
    :members:

.. autoclass:: Serialiser
    :show-inheritance:

.. autoclass:: Deserialiser
    :show-inheritance:

.. autoclass:: MonitoredSerialiser
    :show-inheritance:

.. autoclass:: MonitoredDeserialiser
    :show-inheritance:

.. autofunction:: context_type
"""

from vc2_conformance.py2x_compat import wraps

from contextlib import contextmanager

from vc2_conformance.bitstream.exceptions import (
    UnusedTargetError,
    ReusedTargetError,
    ListTargetExhaustedError,
    ListTargetContainsNonListError,
    UnclosedBoundedBlockError,
    UnclosedNestedContextError,
)


__all__ = [
    "SerDes",
    "Serialiser",
    "Deserialiser",
    "MonitoredSerialiser",
    "MonitoredDeserialiser",
    "context_type",
]


class SerDes(object):
    """
    The base serialiser/deserialiser interface and implementation.

    This base implementation includes all but the value writing/reading
    features of the serialisation and deserialisation process.

    Attributes
    ==========
    io : :py:class:`~.io.BitstreamReader` or :py:class:`~.io.BitstreamWriter`
        The I/O interface in use.
    context : dict or None
        The (top-level) context dictionary.
    cur_context : dict or None
        The context dictionary currently being populated.
    """

    def __init__(self, io, context=None):
        """
        Parameters
        ==========
        io : :py:class:`~.io.BitstreamReader` or :py:class:`~.io.BitstreamWriter`
            The current I/O interface to use initially.
        context : dict
            The initial context dictionary.
        """
        self.io = io

        # The current context dictionary.
        # {target_name: value, ...}
        self.cur_context = context if context is not None else {}

        # Logs which target names have already been used. Initially an empty
        # dictionary.
        #
        # When a non-list target is used, a corresponding entry is set to True in
        # this dictionary. Re-use of a target is prevented by checking that the
        # required target does not already appear in this dictionary.
        #
        # When :py:attr:`declare_list` is used, the corresponding entry in
        # context_indices is set to to 0. Subsequent uses of that target should
        # increment the counter.
        #
        # {target_name: True or int, ...}
        self._cur_context_indices = {}

        # Whenever :py:meth:`subcontext_enter` is used, the current
        # self.cur_context and self._cur_context_indices dictionaries and the
        # specified target name are pushed onto their respective stacks. The
        # :py:meth:`subcontext_leave` method pops these values again.
        self._context_stack = []  # [<context dict>, ...]
        self._context_indices_stack = []  # [<context_indices dict>, ...]
        self._target_stack = []  # [str, ...]

    def _set_context_value(self, target, value):
        """
        Add a value to a context dictionary, checking that the value has not
        already been set and extending list targets if necessary.
        """
        if target not in self._cur_context_indices:
            # Case: This target has not been declared as a list and this is the
            # first time it has been accessed.
            self.cur_context[target] = value
            self._cur_context_indices[target] = True
        elif self._cur_context_indices[target] is True:
            # Case: This target has not been declared as a list and has already
            # been accessed.
            raise ReusedTargetError(self.describe_path(target))
        else:
            # Case: This target has been declared as a list.
            i = self._cur_context_indices[target]
            self._cur_context_indices[target] += 1

            target_list = self.cur_context[target]
            if len(target_list) == i:
                # List is being filled for the first time
                target_list.append(value)
            else:
                # List already exists and we're updating it.
                target_list[i] = value

    def _get_context_value(self, target):
        """
        Get a value from the context dictionary, checking that the value has not
        already been accessed and moving on to the next list item for list
        targets.

        This method may be overridden to modify values fetched from the context
        dictionary (or to choose default values if a suitable value is
        missing).
        """
        if target not in self._cur_context_indices:
            # Case: This target is not a list and has not been used before
            self._cur_context_indices[target] = True
            return self.cur_context[target]
        elif self._cur_context_indices[target] is True:
            # Case: This target is not a list but has already been used
            raise ReusedTargetError(self.describe_path(target))
        else:
            # Case: This target has been declared as a list.
            i = self._cur_context_indices[target]
            if i < len(self.cur_context[target]):
                self._cur_context_indices[target] += 1
                return self.cur_context[target][i]
            else:
                raise ListTargetExhaustedError(self.describe_path(target))

    def _setdefault_context_value(self, target, default):
        """
        Attempt to get a value (or next value, for lists) for a particular
        target. If the value does not exist, sets it to the supplied default.
        """
        if target not in self._cur_context_indices:
            # Case: This target is not a list and has not been used before
            self._cur_context_indices[target] = True
            return self.cur_context.setdefault(target, default)
        elif self._cur_context_indices[target] is True:
            # Case: This target is not a list but has already been used. Fail.
            raise ReusedTargetError(self.describe_path(target))
        else:
            # Case: This target has been declared as a list.
            i = self._cur_context_indices[target]
            self._cur_context_indices[target] += 1
            if i == len(self.cur_context[target]):
                self.cur_context[target].append(default)
            return self.cur_context[target][i]

    def bool(self, target):
        """
        Reads or writes a boolean (single bit) in a bitstream (as per (A.3.2)
        read_bool()).

        Parameters
        ==========
        target : str
            The target for the bit (as a :py:class:`bool`).

        Returns
        =======
        value : bool
        """
        raise NotImplementedError()

    def nbits(self, target, num_bits):
        """
        Reads or writes a fixed-length unsigned integer in a bitstream (as
        per (A.3.3) read_nbits()).

        Parameters
        ==========
        target : str
            The target for the value (as an :py:class:`int`).
        num_bits : int
            The number of bits in the value.

        Returns
        =======
        value : int
        """
        raise NotImplementedError()

    def uint_lit(self, target, num_bytes):
        """
        Reads or writes a fixed-length unsigned integer in a bitstream (as
        per (A.3.4) read_uint_lit()). Not to be confused with :py:meth:`uint`.

        Parameters
        ==========
        target : str
            The target for the value (as an :py:class:`int`).
        num_bytes : int
            The number of bytes in the value.

        Returns
        =======
        value : int
        """
        raise NotImplementedError()

    def bitarray(self, target, num_bits):
        """
        Reads or writes a fixed-length string of bits from the bitstream as a
        :py:class:`bitarray.bitarray`. This may be a more sensible type for
        holding unpredictably sized non-integer binary values such as padding
        bits.

        Parameters
        ==========
        target : str
            The target for the value (as a :py:class:`bitarray.bitarray`).
        num_bits : int
            The number of bits in the value.

        Returns
        =======
        value : :py:class:`bitarray.bitarray`
        """
        raise NotImplementedError()

    def bytes(self, target, num_bytes):
        """
        Reads or writes a fixed-length :py:class:`bytes` string from the
        bitstream. This is a more convenient alternative to :py:meth:`nbits` or
        :py:meth:`bitarray` when large blocks of data are to be read but not
        treated as integers.

        Parameters
        ==========
        target : str
            The target for the value (as a :py:class:`bytes`).
        num_bits : int
            The number of *bytes* (not bits) in the value.

        Returns
        =======
        value : :py:class:`bytes`
        """
        raise NotImplementedError()

    def uint(self, target):
        """
        A variable-length, unsigned exp-golomb integer in a bitstream (as per
        (A.4.3) read_uint()).

        Parameters
        ==========
        target : str
            The target for the value (as an :py:class:`int`).

        Returns
        =======
        value : int
        """
        raise NotImplementedError()

    def sint(self, target, num_bits):
        """
        A variable-length, signed exp-golomb integer in a bitstream (as per (A.4.4)
        read_sint()).

        Parameters
        ==========
        target : str
            The target for the value (as an :py:class:`int`).

        Returns
        =======
        value : int
        """
        raise NotImplementedError()

    def byte_align(self, target):
        """
        Advance in the bitstream to the next whole byte boundary, if not already on
        one (as per (A.2.4) byte_align()).

        Parameters
        ==========
        target : str
            The target for the padding bits (as a
            :py:class:`bitarray.bitarray`).
        """
        _, bits = self.io.tell()
        num_bits = 0 if bits == 7 else (bits + 1)
        self.bitarray(target, num_bits)

    def bounded_block_begin(self, length):
        """
        Defines the start of a bounded block (as per (A.4.2)). Must be followed
        by a matching :py:meth:`bounded_block_end`.

        See also: :py:meth:`bounded_block`.

        Bits beyond the end of the block are always '1'. If a '0' is written
        past the end of the block a :py:exc:`ValueError` will be thrown.

        Parameters
        ==========
        length : int
            The length of the bounded block in bits
        """
        self.io.bounded_block_begin(length)

    def bounded_block_end(self, target):
        """
        Defines the end of a bounded block (as per (A.4.2)). Must be proceeded
        by a matching :py:meth:`bounded_block_begin`.

        Parameters
        ==========
        target : str
            The target name for any unused bits (as a
            :py:class:`bitarray.bitarray`).
        """
        num_bits = self.io.bounded_block_end()
        self.bitarray(target, num_bits)

    @contextmanager
    def bounded_block(self, target, length):
        """
        A context manager defining a bounded block (as per (A.4.2)).

        See also: :py:meth:`bounded_block_begin`.

        Example usage::

            with serdes.bounded_block("unused_bits", 100):
                # ...

        Parameters
        ==========
        target : str
            The target name for any unused bits (as a
            :py:class:`bitarray.bitarray`).
        length : int
            The length of the bounded block in bits
        """
        self.bounded_block_begin(length)
        yield
        self.bounded_block_end(target)

    def declare_list(self, target):
        """
        Declares that the specified target should be treated as a
        :py:class:`list`.  Whenever this target is used in the future, values
        will be read/written sequentially from the list.

        This method has no impact on the bitstream.

        Parameters
        ==========
        target : str
            The target name to be declared as a list.
        """
        if target in self._cur_context_indices:
            # Target has already been used or delcared
            raise ReusedTargetError(self.describe_path(target))

        if target not in self.cur_context:
            # Target not yet defined in context; create a new empty list
            self.cur_context[target] = []
        else:
            # The target already exists in the context; make sure it is a list
            if not isinstance(self.cur_context[target], list):
                raise ListTargetContainsNonListError(
                    "{} contains {!r} (which is not a list)".format(
                        self.describe_path(target), self.cur_context[target]
                    )
                )

        self._cur_context_indices[target] = 0

    def set_context_type(self, context_type):
        """
        Set (or change) the type of the current context dictionary.

        This method has no impact on the bitstream.

        Parameters
        ==========
        context_type : :py:class:`dict`-like type
            The desired type. If the context is already of the required type,
            no change will be made. If the context is currently of a different
            type, it will be passed to the ``context_type`` constructor and the
            new type used in its place.
        """
        # Only replace the type if necessary to avoid unnecessary copying.
        if type(self.cur_context) is not context_type:
            self.cur_context = context_type(self.cur_context)

            # Replace the reference to this context in its parent context
            if self._context_stack:
                parent_context = self._context_stack[-1]
                parent_target = self._target_stack[-1]
                parent_target_index = self._context_indices_stack[-1][parent_target]

                if parent_target_index is True:
                    # The child context is in a normal target in the parent context
                    # dict
                    parent_context[parent_target] = self.cur_context
                else:
                    # The child context is in a list target in the parent context dict
                    # (NB: The parent_target_index value is the *next* index in the
                    # list, hence being decremented by one here).
                    parent_context[parent_target][
                        parent_target_index - 1
                    ] = self.cur_context
            else:
                # Context stack is empty so there must be no parent to update!
                pass

    def subcontext_enter(self, target):
        """
        Creates and/or enters a context dictionary within the specified
        target of the current context dictionary. Must be followed later by a
        matching :py:meth:`subcontext_leave`.

        Parameters
        ==========
        target : str
            The name of the target in the current context in which the new
            subcontext is/will be stored.
        """
        # Insert the new context into the current context at the
        # specified token
        new_context = self._setdefault_context_value(target, {})

        # Push the old context onto the stack
        self._context_stack.append(self.cur_context)
        self._context_indices_stack.append(self._cur_context_indices)
        self._target_stack.append(target)

        self.cur_context = new_context
        self._cur_context_indices = {}

    def subcontext_leave(self):
        """
        Leaves the current nested context dictionary entered by
        :py:meth:`subcontext_enter`. Verifies that the closed dictionary has no
        unused entries, throwing an appropriate exception if not.
        """
        self._verify_context_is_complete()

        self.cur_context = self._context_stack.pop()
        self._cur_context_indices = self._context_indices_stack.pop()
        self._target_stack.pop()

    @contextmanager
    def subcontext(self, target):
        """
        A Python context manager alternative to ;py:meth:`subcontext_enter` and
        ;py:meth:`subcontext_leave`.

        Example usage::

            >>> with serdes.subcontext("target"):
            ...     # ...

        Exactly equivalent to::

            >>> serdes.subcontext_enter("target"):
            >>> # ...
            >>> serdes.subcontext_leave():

        (But without the possibility of forgetting the
        :py:meth:`subcontext_leave` call).

        Parameters
        ==========
        target : str
            The name of the target in the current context in which the new
            subcontext is/will be stored.
        """
        self.subcontext_enter(target)
        yield
        self.subcontext_leave()

    def computed_value(self, target, value):
        """
        Places a value into the named target in the current context, without
        reading or writing anything into the bitstream. Any existing value in
        the context will be overwritten.

        This operation should be used sparingly to embed additional information
        in a context dictionary which might be required to sensibly interpret
        its contents and which cannot be trivially computed from the context
        dictionary. For example, the number of transform coefficients in a
        coded picture depends on numerous computations and table lookups using
        earlier bitstream values.

        Parameters
        ==========
        target : str
            The name of the target in the current context to store the value in.
        value : any
            The value to be stored.
        """
        self._set_context_value(target, value)

    def _verify_target_complete(self, target):
        """
        Verify that a named target in the current context has been
        (completely) used. Raises :py:exc:`UnusedTargetError` if not.
        """
        # Target not used at all
        if target not in self._cur_context_indices:
            raise UnusedTargetError(self.describe_path(target))

        index = self._cur_context_indices[target]
        value = self.cur_context[target]
        if index is True:
            # Target was used (and it is not a list)
            pass
        elif index != len(value):
            # Target was used and was a list, but not all entries were
            # used!
            raise UnusedTargetError(
                "{}[{!r}][{}:{}]".format(
                    self.describe_path(), target, index, len(value)
                )
            )

    def is_target_complete(self, target):
        """
        Test whether a target in the current context is complete, i.e. has been
        fully used up. Returns True if so, False otherwise.
        """
        try:
            self._verify_target_complete(target)
            return True
        except UnusedTargetError:
            return False

    def _verify_context_is_complete(self):
        """
        Verify that the current context is 'complete'. That is, every entry in
        the context dict has been used and that every element in any lists has
        been used too.

        Raises
        ======
        :py:exc:`~.UnusedTargetError`
        """
        for target in self.cur_context:
            self._verify_target_complete(target)

    def verify_complete(self):
        """
        Verify that all values in the current context have been used and that
        no bounded blocks or nested contexts have been left over.

        Raises
        ======
        :py:exc:`~.UnusedTargetError`
        :py:exc:`~.UnclosedNestedContextError`
        :py:exc:`~.UnclosedBoundedBlockError`
        """
        self._verify_context_is_complete()

        if self._context_stack:
            raise UnclosedNestedContextError(self.describe_path())

        if self.io.bits_remaining is not None:
            raise UnclosedBoundedBlockError()

    def __enter__(self):
        """
        When used as a context manager, 'verify_complete' is automatically
        called.

        Example usage::

            >>> with Deserialiser(reader) as serdes:
            ...     frame_size(serdes)
            >>> serdes.context

        Exactly equivalent to::

            >>> serdes = Deserialiser(reader)
            >>> frame_size(serdes)
            >>> serdes.verify_complete()
            >>> serdes.context

        (But without the possibility of foregoing the
        :py:meth:`verify_complete` call.)
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Don't bother checking validity if an exception was thrown anyway
        if exc_type is None:
            self.verify_complete()

    @property
    def context(self):
        """Get the top-level context dictionary."""
        if self._context_stack:
            return self._context_stack[0]
        else:
            return self.cur_context

    def path(self, target=None):
        """
        Produce a 'path' describing the part of the bitstream the parser is
        currently processing.

        If 'target' is None, only includes the path of the current nested context
        dictionary. If 'target' is a target name in the current target
        dictionary, the path to the last-used target will be included.

        A path might look like::

            ['source_parameters', 'frame_size', 'frame_width']
        """

        full_context_stack = list(self._context_stack)
        full_context_indices_stack = list(self._context_indices_stack)
        full_target_stack = list(self._target_stack)
        if target is not None:
            full_context_stack += [self.cur_context]
            full_context_indices_stack += [self._cur_context_indices]
            full_target_stack += [target]

        out = []

        for context, context_indices, target in zip(
            full_context_stack, full_context_indices_stack, full_target_stack
        ):
            out.append(target)

            index = context_indices.get(target, True)
            if index is not True and index != 0:
                # NB: context_indices includes the index of the first unused
                # index, hence being decremented by one here.
                out.append(index - 1)

        return out

    def describe_path(self, target=None):
        """
        Produce a human-readable description of the part of the bitstream the
        parser is currently processing.

        If 'target' is None, prints only the path of the current nested context
        dictionary. If 'target' is a target name in the current target
        dictionary, this will be included in the string.

        As a sample, a path might look like the following::

            SequenceHeader['source_parameters']['frame_size']['frame_width']

        """
        root_type = self.context.__class__.__name__

        return "{}{}".format(
            root_type, "".join("[{!r}]".format(p) for p in self.path(target))
        )


class Deserialiser(SerDes):
    """
    A bitstream deserialiser which creates a context dictionary based on a
    bitstream.

    Parameters
    ==========
    io : :py:class:`~.io.BitstreamReader`
    """

    def bool(self, target):
        value = bool(self.io.read_bit())
        self._set_context_value(target, value)
        return value

    def nbits(self, target, num_bits):
        value = self.io.read_nbits(num_bits)
        self._set_context_value(target, value)
        return value

    def uint_lit(self, target, num_bytes):
        value = self.io.read_uint_lit(num_bytes)
        self._set_context_value(target, value)
        return value

    def bitarray(self, target, num_bits):
        value = self.io.read_bitarray(num_bits)
        self._set_context_value(target, value)
        return value

    def bytes(self, target, num_bytes):
        value = self.io.read_bytes(num_bytes)
        self._set_context_value(target, value)
        return value

    def uint(self, target):
        value = self.io.read_uint()
        self._set_context_value(target, value)
        return value

    def sint(self, target):
        value = self.io.read_sint()
        self._set_context_value(target, value)
        return value


class Serialiser(SerDes):
    """
    A bitstream serialiser which, given a populated context dictionary, writes
    the corresponding bitstream.

    Parameters
    ==========
    io : :py:class:`~.io.BitstreamWriter`
    context : dict
    """

    def __init__(self, io, context=None, default_values={}):
        super(Serialiser, self).__init__(io, context)
        self.default_values = default_values

    def _get_context_value(self, target):
        """
        Get a value from the context dictionary, checking that the value has not
        already been accessed and moving on to the next list item for list
        targets.
        """
        try:
            return super(Serialiser, self)._get_context_value(target)
        except (KeyError, ListTargetExhaustedError):
            # Fall back on default value if provided
            context_type = type(self.cur_context)
            if (
                context_type in self.default_values
                and target in self.default_values[context_type]
            ):
                return self.default_values[context_type][target]
            else:
                raise

    def bool(self, target):
        value = self._get_context_value(target)
        self.io.write_bit(value)
        return bool(value)

    def nbits(self, target, num_bits):
        value = self._get_context_value(target)
        self.io.write_nbits(num_bits, value)
        return value

    def uint_lit(self, target, num_bytes):
        value = self._get_context_value(target)
        self.io.write_uint_lit(num_bytes, value)
        return value

    def bitarray(self, target, num_bits):
        value = self._get_context_value(target)
        self.io.write_bitarray(num_bits, value)
        return value

    def bytes(self, target, num_bytes):
        value = self._get_context_value(target)
        self.io.write_bytes(num_bytes, value)
        return value

    def uint(self, target):
        value = self._get_context_value(target)
        self.io.write_uint(value)
        return value

    def sint(self, target):
        value = self._get_context_value(target)
        self.io.write_sint(value)
        return value


class MonitoredMixin(object):
    """
    A mixin for :py:class:`SerDes` classes which allows a 'monitor' function to
    be provided which will be called after each primitive I/O operation
    completes. This allows the intermediate progress of bistream
    serialisation/deserialisation to be monitored or even terminated early
    (with an exception).
    """

    def __init__(self, monitor, *args, **kwargs):
        """
        Parameters
        ==========
        monitor : callable(serdes, target, value)
            A function which will be called after every primitive I/O operation
            completes. This function is passed the :py:class:`SerDes` instance
            and the target name and value of the target just used.
        *args, **kwargs :
            Passed to base constructor.
        """
        super(MonitoredMixin, self).__init__(*args, **kwargs)
        self.monitor = monitor

    def bool(self, target):
        value = super(MonitoredMixin, self).bool(target)
        self.monitor(self, target, value)
        return value

    def nbits(self, target, num_bits):
        value = super(MonitoredMixin, self).nbits(target, num_bits)
        self.monitor(self, target, value)
        return value

    def uint_lit(self, target, num_bytes):
        value = super(MonitoredMixin, self).uint_lit(target, num_bytes)
        self.monitor(self, target, value)
        return value

    def bitarray(self, target, num_bits):
        value = super(MonitoredMixin, self).bitarray(target, num_bits)
        self.monitor(self, target, value)
        return value

    def bytes(self, target, num_bytes):
        value = super(MonitoredMixin, self).bytes(target, num_bytes)
        self.monitor(self, target, value)
        return value

    def uint(self, target):
        value = super(MonitoredMixin, self).uint(target)
        self.monitor(self, target, value)
        return value

    def sint(self, target):
        value = super(MonitoredMixin, self).sint(target)
        self.monitor(self, target, value)
        return value


class MonitoredSerialiser(MonitoredMixin, Serialiser):
    """
    Like :py:class:`Serialiser` but takes a 'monitor' function as the first
    constructor argument. This function will be called every time bitstream
    value has been serialised (written).

    Parameters
    ==========
    monitor : callable(ser, target, value)
        A function which will be called after every primitive I/O operation
        completes. This function is passed this :py:class:`MonitoredSerialiser`
        instance and the target name and value of the target just serialised.

        This function may be used to inform a user of the current progress of
        serialisation (e.g. using :py:meth:`SerDes.describe_path` or
        :py:data:`SerDes.io`) or to terminate serialisation early (by throwing
        an exception).
    *args, **kwargs : see :py:class:`Serialiser`
    """


class MonitoredDeserialiser(MonitoredMixin, Deserialiser):
    """
    Like :py:class:`Deserialiser` but takes a 'monitor' function as the first
    constructor argument. This function will be called every time bitstream
    value has been deserialised (read).

    Parameters
    ==========
    monitor : callable(des, target, value)
        A function which will be called after every primitive I/O operation
        completes. This function is passed this
        :py:class:`MonitoredDeserialiser` instance and the target name and
        value of the target just serialised.

        This function may be used to inform a user of the current progress of
        deserialisation (e.g. using :py:meth:`SerDes.context`,
        :py:meth:`SerDes.describe_path` or :py:data:`SerDes.io`) or to
        terminate deserialisation early (by throwing an exception).
    *args, **kwargs : see :py:class:`Serialiser`
    """


def context_type(dict_type):
    """
    Syntactic sugar. A decorator for :py:class:`SerDes` which uses
    :py:meth:`SerDes.set_context_type` to set the type of the current context
    dict:

    Example usage::

        @context_type(FrameSize)
        def frame_size(serdes):
            # ...

    Exactly equivalent to::

        def frame_size(serdes):
            serdes.set_context_type(FrameSize)
            # ...

    The wrapped function must take a :py:class:`SerDes` as its first argument.

    For introspection purposes, the wrapper function will be given a
    'context_type' attribute holding the passed 'dict_type'.
    """

    def wrap(f):
        @wraps(f)
        def wrapper(serdes, *args, **kwargs):
            serdes.set_context_type(dict_type)
            return f(serdes, *args, **kwargs)

        wrapper.context_type = dict_type
        wrapper.__wrapped__ = f
        return wrapper

    return wrap



from vc2_data_tables import ParseCodes

from vc2_conformance.bitstream import Stream

from vc2_conformance.test_cases import decoder_test_case_generator

from vc2_conformance.encoder import make_sequence

from vc2_conformance.picture_generators import (
    mid_gray,
    repeat_pictures,
)


@decoder_test_case_generator
def absent_next_parse_offset(codec_features):
    """
    **Tests handling of missing 'next parse offset' field.**

    The 'next parse offset' field of the ``parse_info`` header (see (10.5.1))
    can be set to zero (i.e. omitted) for pictures. This test case verifies
    that decoders are still able to decode streams with this field absent.
    """
    sequence = make_sequence(
        codec_features,
        repeat_pictures(
            mid_gray(
                codec_features["video_parameters"],
                codec_features["picture_coding_mode"],
            ),
            2,
        ),
    )

    # Prevent the default auto numbering of picture-containing data units
    # during serialisation
    for data_unit in sequence["data_units"]:
        parse_info = data_unit["parse_info"]
        if parse_info["parse_code"] in (
            ParseCodes.low_delay_picture,
            ParseCodes.high_quality_picture,
            ParseCodes.low_delay_picture_fragment,
            ParseCodes.high_quality_picture_fragment,
        ):
            parse_info["next_parse_offset"] = 0

    return Stream(sequences=[sequence])



"""
The :py:mod:`vc2_conformance.dimensions_and_depths` module contains the
:py:func:`compute_dimensions_and_depths` function which returns the picture
dimensions and bit depths implied by a VC-2 video format.

.. autofunction:: compute_dimensions_and_depths

.. autoclass:: DimensionsAndDepths

"""

from collections import OrderedDict, namedtuple

from vc2_conformance.pseudocode.vc2_math import intlog2

from vc2_conformance.pseudocode.state import State

from vc2_conformance.pseudocode.video_parameters import set_coding_parameters


__all__ = [
    "DimensionsAndDepths",
    "compute_dimensions_and_depths",
]


DimensionsAndDepths = namedtuple(
    "DimensionsAndDepths",
    "width,height,depth_bits,bytes_per_sample",
)
"""
A set of picture component dimensions and bit depths.

Parameters
==========
width, height : int
    The dimensions of the picture.
depth_bits : int
    The number of bits per pixel.
bytes_per_sample : int
    The number of bytes used to store each pixel value in raw file formats (see
    :py:mod:`vc2_conformance.file_format`).
"""


def compute_dimensions_and_depths(video_parameters, picture_coding_mode):
    """
    Compute the dimensions, bit depth and bytes-per-sample of a picture.

    Parameters
    ==========
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`

    Returns
    =======
    OrderedDict
        An ordered dictionary mapping from component name ("Y", "C1" and "C2")
        to a :py:class:`DimensionsAndDepths` (width, height, depth_bits,
        bytes_per_sample) namedtuple.
    """
    # NB: We use the pseudocode to actually compute the dimensions and signal
    # range to avoid mistakes
    state = State(picture_coding_mode=picture_coding_mode)
    set_coding_parameters(state, video_parameters)

    out = OrderedDict()

    for component in ["Y", "C1", "C2"]:
        if component == "Y":
            width = state["luma_width"]
            height = state["luma_height"]
            depth_bits = state["luma_depth"]
        else:
            width = state["color_diff_width"]
            height = state["color_diff_height"]
            depth_bits = state["color_diff_depth"]

        bytes_per_sample = (depth_bits + 7) // 8  # Round up to whole number of bytes
        bytes_per_sample = 1 << intlog2(bytes_per_sample)  # Round up to power of two

        out[component] = DimensionsAndDepths(
            width, height, depth_bits, bytes_per_sample
        )

    return out



"""
Test specifically for lossless decoders that verifies basic support for
quantisation.
"""

import logging

from copy import deepcopy

from io import BytesIO

from vc2_conformance.bitstream import Stream, autofill_and_serialise_stream

from vc2_conformance.test_cases import decoder_test_case_generator

from vc2_conformance.picture_generators import mid_gray

from vc2_conformance.encoder import make_sequence

from vc2_conformance.pseudocode.state import State

from vc2_conformance.test_cases.decoder.common import (
    iter_transform_parameters_in_sequence,
    iter_slices_in_sequence,
)

from vc2_conformance.encoder.pictures import (
    get_quantization_marix,
    calculate_hq_length_field,
)

from vc2_conformance.decoder import init_io, parse_stream


MINIMUM_DISTINCT_QINDEX = 7
"""
The quantization index at and above which the value of ``inverse_quant(1, qi)``
is distinct for all quantization indices.
"""


def compute_qindex_with_distinct_quant_factors(quant_matrix):
    """
    Given a quantization matrix, return a qindex high enough that all
    quantization factors used will be greater than
    :py:data:`MINIMUM_DISTINCT_QINDEX`.
    """
    # The quantization index should be chosen such that:
    #
    # * Even for the most heavily skewed quantization matrix entry some
    #   quantisation still occurs
    # * When inverse quantisation is applied to '1', the result should be
    #   distinct for all quantisation indices used
    max_quant_matrix_value = max(
        v for subbands in quant_matrix.values() for v in subbands.values()
    )
    return max_quant_matrix_value + MINIMUM_DISTINCT_QINDEX


def check_for_signal_clipping(sequence):
    """
    Given a :py:class:`vc2_conformance.bitstream.Sequence`, return True if any
    picture component signal was clipped during decoding.
    """
    # NB: Internally we just check for saturated signal levels. This way we
    # avoid the need to modify the decoder to remove the clipper and all that
    # faff...

    # Serialise
    f = BytesIO()
    # NB: Deepcopy required due to autofill_and_serialise_stream mutating the
    # stream
    stream = Stream(sequences=[deepcopy(sequence)])
    autofill_and_serialise_stream(f, stream)
    f.seek(0)

    # Decode and look for saturated pixel values
    state = State()
    may_have_clipped = [False]

    def output_picture_callback(picture, video_parameters, picture_coding_mode):
        components_and_depths = [
            ("Y", state["luma_depth"]),
            ("C1", state["color_diff_depth"]),
            ("C2", state["color_diff_depth"]),
        ]

        for component, depth in components_and_depths:
            min_value = min(min(row) for row in picture[component])
            max_value = max(max(row) for row in picture[component])
            if min_value == 0:
                may_have_clipped[0] = True
            if max_value == (1 << depth) - 1:
                may_have_clipped[0] = True

    state["_output_picture_callback"] = output_picture_callback
    init_io(state, f)
    parse_stream(state)

    return may_have_clipped[0]


@decoder_test_case_generator
def lossless_quantization(codec_features):
    """
    **Tests support for quantization in lossless decoders.**

    Quantization can, in principle, be used in lossless coding modes in cases
    where all transform coefficients are divisible by the same factor. This
    test case contains a synthetic test pattern with this property.

    This test case is only generated for lossless codecs.

    .. note::

        For informational purposes, an example decoded test pattern is shown
        below:

        .. image:: /_static/user_guide/lossless_quantization.png

        Note the faint repeating pattern.
    """
    # Don't bother with this test for lossy coding modes (quantization is
    # tested elsewhere)
    if not codec_features["lossless"]:
        return None

    # Pick a non-zero qindex which will ensure all transform coefficients, when
    # set to 1 in the bitstream, will dequantize to different values (when the
    # quant matrix entry is different).
    quant_matrix = get_quantization_marix(codec_features)
    qindex = compute_qindex_with_distinct_quant_factors(quant_matrix)

    # Start with a mid-gray frame (coeffs set to 0). We'll hand-modify this to
    # contain all 1s because a picture which does this may be slightly larger
    # than the unclipped picture size and therefore we can't rely on the
    # encoder to produce such a signal.
    sequence = make_sequence(
        codec_features,
        mid_gray(
            codec_features["video_parameters"],
            codec_features["picture_coding_mode"],
        ),
    )

    # Set qindex and all transform coefficients to 1
    max_length = 0
    for _state, _sx, _sy, hq_slice in iter_slices_in_sequence(codec_features, sequence):
        hq_slice["qindex"] = qindex
        for c in ["y", "c1", "c2"]:
            hq_slice["{}_transform".format(c)] = [
                1 for _ in hq_slice["{}_transform".format(c)]
            ]
            length = calculate_hq_length_field(hq_slice["{}_transform".format(c)], 1)
            hq_slice["slice_{}_length".format(c)] = length
            max_length = max(length, max_length)

    # Update slice size scaler to keep all length fields to 8 bits or fewer
    slice_size_scaler = max(1, (max_length + 254) // 255)
    for transform_parameters in iter_transform_parameters_in_sequence(
        codec_features, sequence
    ):
        transform_parameters["slice_parameters"][
            "slice_size_scaler"
        ] = slice_size_scaler
    for _state, _sx, _sy, hq_slice in iter_slices_in_sequence(codec_features, sequence):
        for c in ["y", "c1", "c2"]:
            hq_slice["slice_{}_length".format(c)] += slice_size_scaler - 1
            hq_slice["slice_{}_length".format(c)] //= slice_size_scaler

    # If the resulting picture clips, give up on this test case. We make the
    # assumption that while a clever lossless encoder may use quantization it
    # is unlikely to rely on signal clipping in the decoder. As a consequence,
    # to avoid producing a test case which a decoder might reasonably fail to
    # decode due to internal signal width limitations, we bail.
    #
    # In practice, even for the largest VC-2 filters, transform depths and
    # wonkiest quantisation matrices, the generated signals should fit (very)
    # comfortably into 8 bit video signal ranges. As such, if this check fails
    # it is very likely a highly degenerate codec configuration has been
    # specified.
    if check_for_signal_clipping(sequence):
        logging.warning(
            "The lossless_quantization test case generator could not produce a "
            "losslessly compressible image and has been omitted. This probably "
            "means an (impractically) high transform depth or custom quantisation "
            "matrix entry or an extremely low picture bit depth was used."
        )
        return None

    return Stream(sequences=[sequence])



"""
Generic synthetic test sequences.
"""

from vc2_conformance.test_cases import encoder_test_case_generator

from vc2_conformance.test_cases.encoder.common import picture_generator_to_test_case

from vc2_conformance import picture_generators


@encoder_test_case_generator
def synthetic_moving_sprite(codec_features):
    """
    **Tests that an encoder produces sensible results for motion.**

    A sequence of 10 frames containing a graphic moving from left to right
    along the top of the frame. In successive each frame, the graphic moves
    16 luma samples to the right (i.e. 8 samples every field, for
    interlaced formats).

    .. image:: /_static/user_guide/interlace_mode_and_pixel_aspect_ratio_moving_sequence.svg

    For progressive formats, the graphic should appear with smooth edges in
    each frame.

    For interlaced formats, the graphic should move smoothly when displayed
    on an interlaced monitor. If displayed as progressive frames (as in the
    illustration above), the pictures will appear to have ragged edges.
    """
    return picture_generator_to_test_case(
        picture_generators.moving_sprite,
        codec_features,
    )


@encoder_test_case_generator
def synthetic_linear_ramps(codec_features):
    """
    **Tests that an encoder correctly encodes color specification
    information.**

    A static frame containing linear signal ramps for white and primary
    red, green and blue (in that order, from top-to-bottom) as illustrated
    below:

    .. image:: /_static/user_guide/static_ramps.png

    The red, green and blue colors correspond to the red, green and blue
    primaries for the color specification (11.4.10.2).

    .. note::

        When D-Cinema primaries are specified (preset color primaries index 3),
        red, green and blue are replaced with CIE X, Y and Z respectively. Note
        that these might not represent physically realisable colors.

    The left-most pixels in each band are video black and the right-most pixels
    video white, red, green and blue (respectively). That is, oversaturated
    signals (e.g. 'super-blacks' and 'super-white') are not included.

    The value ramps in the test picture are linear meaning that the (linear)
    pixel values increase at a constant rate from left (black) to right
    (saturated white/red/green/blue). Due to the non-linear response of human
    vision, this will produce a non-linear brightness ramp which appears to
    quickly saturate. Further, when a non-linear transfer function is specified
    (11.4.10.4) the raw picture values will not be linearly spaced.

    .. note::

        When the D-Cinema transfer function is specified (preset transfer
        function index 3), the saturated signals do not correspond to a
        non-linear signal value of 1.0 but instead approximately 0.97. This is
        because the D-Cinema transfer function allocates part of its nominal
        output range to over-saturated signals.
    """
    return picture_generator_to_test_case(
        picture_generators.linear_ramps,
        codec_features,
    )


@encoder_test_case_generator
def synthetic_gray(codec_features):
    """
    **Tests that the encoder can encode a maximally compressible sequence.**

    This sequence contains an image in which every transform coefficient is
    zero. For most color specifications (11.4.10), this decodes to a mid-gray
    frame.

    This special case image is maximally compressible since no transform
    coefficients need to be explicitly coded in the bitstream. For lossless
    coding modes, this should also produce produce the smallest possible
    bitstream.
    """
    return picture_generator_to_test_case(
        picture_generators.mid_gray,
        codec_features,
    )


@encoder_test_case_generator
def synthetic_noise(codec_features):
    """
    **Tests that an encoder correctly encodes a noise plate.**

    A static frame containing pseudo-random uniform noise as illustrated below:

    .. image:: /_static/user_guide/noise.png

    .. note::

        It is likely that lossy encoders will be unable to compress this test
        case without a fairly significant loss of fidelity. As such, it is
        acceptable for this test case for an encoder to produce only visually
        similar results.
    """
    return picture_generator_to_test_case(
        picture_generators.white_noise,
        codec_features,
    )



r"""
.. _vc2-picture-compare:

``vc2-picture-compare``
=======================

A command-line utility which compares pairs of raw pictures (see
:ref:`file-format`), or pairs of directories containing a series of raw
pictures.

Usage
-----

Given a pair of images in raw format (see :ref:`file-format`), these images can
be compared as follows::

    $ vc2-picture-compare image_a.raw image_b.raw
    Pictures are different:
      Y: Different: PSNR = 55.6 dB, 1426363 pixels (68.8%) differ
      C1: Different: PSNR = 57.7 dB, 662607 pixels (63.9%) differ
      C2: Different: PSNR = 56.8 dB, 703531 pixels (67.9%) differ

Alternatively a pair of directories containing images whose filenames end
with a number (and the extensions ``.raw`` and ``.json``). For example::

    $ vc2-picture-compare ./directory_a/ ./directory_b/
    Comparing ./directory_a/picture_0.raw and ./directory_b/picture_0.raw:
      Pictures are identical
    Comparing ./directory_a/picture_1.raw and ./directory_b/picture_1.raw:
      Pictures are different:
        Y: Different: PSNR = 55.6 dB, 1426363 pixels (68.8%) differ
        C1: Different: PSNR = 57.7 dB, 662607 pixels (63.9%) differ
        C2: Different: PSNR = 56.8 dB, 703531 pixels (67.9%) differ
    Comparing ./directory_a/picture_2.raw and ./directory_b/picture_2.raw:
      Pictures are identical
    Summary: 2 identical, 1 different

Differences in the encoded values are reported separately for each picture
component.

Peak Signal to Noise Ratio (PSNR) figures give the PSNR of the raw integer
signal values (and not, for example, linear light levels).

.. note::

    PSNR is computed as:

    .. math::

        \text{MeanSquareError} &= \sum_{i, j} (E_{i, j} - A_{i, j})^2
        \\
        \text{MaxE} &= \max_{i, j} (E_{i, j})
        \\
        \text{PSNR} &= 20 \log_{10}(\text{MaxE}) - 10 \log_{10}(\text{MeanSquareError})

    Where :math:`E_{i,j}` is pixel :math:`(i, j)` of the raw integer signal
    value of the test image and :math:`A_{i, j}` is pixel :math:`(i, j)` of the
    raw integer values of the decoded image.

The number of pixels which are not bit-for-bit identical is also reported.

The tool also compares the metadata of the raw images and will flag up
differences here too, for example::

    $ vc2-picture-compare picture_hd.raw picture_cif.raw
    Picture numbers are different:
    Video parameters are different:
      - frame_width: 1920
      + frame_width: 352
      - frame_height: 1080
      + frame_height: 288
      - color_diff_format_index: color_4_2_2 (1)
      + color_diff_format_index: color_4_2_0 (2)
        source_sampling: progressive (0)
        top_field_first: True
      - frame_rate_numer: 50
      + frame_rate_numer: 25
      - frame_rate_denom: 1
      + frame_rate_denom: 2
      - pixel_aspect_ratio_numer: 1
      + pixel_aspect_ratio_numer: 12
      - pixel_aspect_ratio_denom: 1
      + pixel_aspect_ratio_denom: 11
      - clean_width: 1920
      + clean_width: 352
      - clean_height: 1080
      + clean_height: 288
        left_offset: 0
        top_offset: 0
      - luma_offset: 64
      + luma_offset: 0
      - luma_excursion: 876
      + luma_excursion: 255
      - color_diff_offset: 512
      + color_diff_offset: 128
      - color_diff_excursion: 896
      + color_diff_excursion: 255
      - color_primaries_index: hdtv (0)
      + color_primaries_index: sdtv_625 (2)
      - color_matrix_index: hdtv (0)
      + color_matrix_index: sdtv (1)
        transfer_function_index: tv_gamma (0)

When picture metadata differs, the pixel values will not be compared since the
difference in metadata typically means a difference in format making the
pictures incomparable.


Generating difference masks
---------------------------

The ``vc2-picture-compare`` tool can optionally output a simple difference mask
image using ``--difference-mask``/``-D``. The generated image contains white
pixels wherever the inputs differed and black pixels wherever they were
identical. The generated difference mask is output as a raw file of the same
format as the two input files. This mode is only supported when two individual
files are provided, not two directories.

For example::

    $ vc2-picture-compare \
          image_a.raw \
          image_b.raw \
          --difference-mask difference_mask.raw
    Pictures are different:
      Y: Different: PSNR = 55.6 dB, 1426363 pixels (68.8%) differ
      C1: Different: PSNR = 57.7 dB, 662607 pixels (63.9%) differ
      C2: Different: PSNR = 56.8 dB, 703531 pixels (67.9%) differ


Arguments
---------

The complete set of arguments can be listed using ``--help``

.. program-output:: vc2-picture-compare --help

"""

import os

import sys

import re

from argparse import ArgumentParser

from collections import OrderedDict

import numpy as np

from vc2_data_tables import ColorDifferenceSamplingFormats

from vc2_conformance import __version__

from vc2_conformance.string_utils import indent

from vc2_conformance.color_conversion import (
    from_xyz,
    matmul_colors,
    LINEAR_RGB_TO_XYZ,
)

from vc2_conformance.file_format import (
    get_metadata_and_picture_filenames,
    read_metadata,
    read_picture,
    write,
)

from vc2_conformance.dimensions_and_depths import compute_dimensions_and_depths

from vc2_conformance.pseudocode.video_parameters import VideoParameters


def read_pictures_with_only_one_metadata_file_required(filename_a, filename_b):
    """
    Read a pair of raw image files. If one of the two is missing its
    corresponding JSON metadata file, the metadata for the other file will be
    used.

    Returns a tuple with the following values:
    * ``(picture_a, video_parameters_a, picture_coding_mode_a)``
    * ``(picture_b, video_parameters_b, picture_coding_mode_b)``
    * ``byte_for_byte_identical``: This indicates that the two raw files
      are byte-for-byte identical, including the bits not responsible for
      encoding images (e.g. when 10-bit values are encoded as 16-bit words,
      this includes a check of the upper 6 bits). This can catch mistakes in
      formatting raw data files.
    """
    meta_fn_a, pic_fn_a = get_metadata_and_picture_filenames(filename_a)
    meta_fn_b, pic_fn_b = get_metadata_and_picture_filenames(filename_b)

    try:
        with open(meta_fn_a, "rb") as f:
            metadata_a = read_metadata(f)
    except (OSError, IOError):
        metadata_a = None

    try:
        with open(meta_fn_b, "rb") as f:
            metadata_b = read_metadata(f)
    except (OSError, IOError):
        metadata_b = None

    if metadata_a is None and metadata_b is None:
        sys.stderr.write("Error: Missing JSON metadata file for both pictures.\n")
        sys.exit(100)
    elif metadata_a is None:
        metadata_a = metadata_b
        sys.stderr.write(
            "Warning: Metadata missing for first picture. "
            "Will assume it is the same as the second.\n"
        )
    elif metadata_b is None:
        metadata_b = metadata_a
        sys.stderr.write(
            "Warning: Metadata missing for second picture. "
            "Will assume it is the same as the first.\n"
        )

    (
        video_parameters_a,
        picture_coding_mode_a,
        picture_number_a,
    ) = metadata_a

    (
        video_parameters_b,
        picture_coding_mode_b,
        picture_number_b,
    ) = metadata_b

    try:
        with open(pic_fn_a, "rb") as f:
            picture_a = read_picture(
                video_parameters_a,
                picture_coding_mode_a,
                picture_number_a,
                f,
            )
            if len(f.read(1)) != 0:
                raise ValueError()
    except (OSError, IOError):
        sys.stderr.write("Error: Could not open first picture.\n")
        sys.exit(101)
    except ValueError:
        sys.stderr.write("Error: First picture file has incorrect size.\n")
        sys.exit(102)

    try:
        with open(pic_fn_b, "rb") as f:
            picture_b = read_picture(
                video_parameters_b,
                picture_coding_mode_b,
                picture_number_b,
                f,
            )
            if len(f.read(1)) != 0:
                raise ValueError()
    except (OSError, IOError):
        sys.stderr.write("Error: Could not open second picture.\n")
        sys.exit(101)
    except ValueError:
        sys.stderr.write("Error: Second picture file has incorrect size.\n")
        sys.exit(102)

    with open(pic_fn_a, "rb") as fa:
        with open(pic_fn_b, "rb") as fb:
            byte_for_byte_identical = fa.read() == fb.read()

    return (
        (picture_a, video_parameters_a, picture_coding_mode_a),
        (picture_b, video_parameters_b, picture_coding_mode_b),
        byte_for_byte_identical,
    )


def parse_args(*args, **kwargs):
    parser = ArgumentParser(
        description="""
        Compare a pair of pictures in the raw format used by the VC-2
        conformance software.
    """
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {}".format(__version__),
    )

    parser.add_argument(
        "filename_a",
        help="""
            The filename of a .raw or .json raw picture, or a directory
            containing a series of .raw and .json files whose names end in a
            number.
        """,
    )

    parser.add_argument(
        "filename_b",
        help="""
            The filename of a .raw or .json raw picture, or a directory
            containing a series of .raw and .json files whose names end in a
            number.
        """,
    )

    output_group = parser.add_argument_group(
        "difference image options",
    )

    output_group.add_argument(
        "--difference-mask",
        "-D",
        type=str,
        metavar="FILENAME",
        help="""
            Output a difference mask image to the specified file. This mask
            will contain white pixels wherever the input images differ and
            black pixels where they match. Only available when comparing files
            (not directories).
        """,
    )

    return parser.parse_args(*args, **kwargs)


def video_parameter_diff(a, b):
    """
    Given a pair of :py:class:`~vc2_conformance.VideoParameters` dictionaries,
    return a diff-style string describing the difference between them.
    """
    # [(prefix, name, value), ...]
    out_lines = []
    for name, entry_obj in VideoParameters.entry_objs.items():
        if name not in a and name not in b:
            continue
        elif name not in b:
            out_lines.append(("-", name, entry_obj.to_string(a[name])))
        elif name not in a:
            out_lines.append(("+", name, entry_obj.to_string(b[name])))
        elif a[name] == b[name]:
            out_lines.append((" ", name, entry_obj.to_string(a[name])))
        else:
            out_lines.append(("-", name, entry_obj.to_string(a[name])))
            out_lines.append(("+", name, entry_obj.to_string(b[name])))

    return "\n".join(
        "{} {}: {}".format(prefix, name, value) for prefix, name, value in out_lines
    )


def picture_coding_mode_diff(a, b):
    """
    Return a string showing the difference between two picture coding modes.
    """
    return "- {} ({:d})\n+ {} ({:d})".format(a.name, a, b.name, b)


def picture_number_diff(a, b):
    """
    Return a string showing the difference between two picture numbers.
    """
    return "- {}\n+ {}".format(a, b)


def psnr(deltas, max_value):
    """
    Compute the peak signal to noise ratio (in dB) given a series of error
    values and associated maximum signal value.

    Returns None if the deltas are zero.
    """
    mean_square_error = np.mean(deltas * deltas)
    if mean_square_error == 0:
        return None
    else:
        return (20 * (np.log(max_value) / np.log(10))) - (
            10 * (np.log(mean_square_error) / np.log(10))
        )


def measure_differences(all_deltas, video_parameters, picture_coding_mode):
    """
    Given a 2D array of pixel value differences for each picture component,
    returns a string summarising the differences.
    """
    dimensions_and_depths = compute_dimensions_and_depths(
        video_parameters,
        picture_coding_mode,
    )

    delta_counts = {c: np.count_nonzero(d) for c, d in all_deltas.items()}
    psnrs = {
        component: psnr(deltas, (1 << dimensions_and_depths[component].depth_bits) - 1)
        for component, deltas in all_deltas.items()
    }

    identical = all(psnr is None for psnr in psnrs.values())

    differences = "\n".join(
        "{}: Identical".format(component)
        if psnrs[component] is None
        else "{}: Different: PSNR = {:.1f} dB, {} pixel{} ({:.1f}%) differ{}".format(
            component,
            psnrs[component],
            delta_counts[component],
            "s" if delta_counts[component] != 1 else "",
            (delta_counts[component] * 100.0) / all_deltas[component].size,
            "s" if delta_counts[component] == 1 else "",
        )
        for component in ["Y", "C1", "C2"]
    )

    return identical, differences


def generate_difference_mask_picture(deltas, video_parameters, picture_number=0):
    """
    Given a set of pixel delta values, produce a difference mask picture (which
    is white where the pixels have a non-zero difference and black otherwise).

    Parameters
    ==========
    deltas : {"Y": np.array, "C1": np.array, "C2": np.array}
        Deltas for each pixel in the input images.
    video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    picture_number : int

    Returns
    =======
    mask_picture : {"Y": [[...]], "C1": [[...]], "C2": [[...]], "pic_num": int}
        A mask picture, in the same format as the original pictures.

        The mask will be white in all pixels for which *any* picture component
        differs at the corresponding pixel location. For formats using
        subsampled color differences, a difference in a color difference
        component will result in 2 or 4 pixels in the mask being illuminated.

        The 'white' and 'black' reported in the mask are video white and video
        black, not super white or super black (for color formats which support
        this).
    """
    masks = {c: deltas[c] != 0 for c, d in deltas.items()}

    # Upsample color difference components
    repeat_x, repeat_y = {
        ColorDifferenceSamplingFormats.color_4_4_4: (1, 1),
        ColorDifferenceSamplingFormats.color_4_2_2: (2, 1),
        ColorDifferenceSamplingFormats.color_4_2_0: (2, 2),
    }[video_parameters["color_diff_format_index"]]

    for c in ["C1", "C2"]:
        if repeat_x != 1:
            masks[c] = np.repeat(masks[c], repeat_x, axis=1)
        if repeat_y != 1:
            masks[c] = np.repeat(masks[c], repeat_y, axis=0)

    # Create the mask
    mask = masks["Y"] | masks["C1"] | masks["C2"]

    # Convert to RGB color image with signals in the range 0.0 - 1.0
    rgb = np.repeat(mask, 3).reshape(mask.shape + (3,)).astype(float)

    # Convert to native color encoding
    xyz = matmul_colors(
        LINEAR_RGB_TO_XYZ[video_parameters["color_primaries_index"]], rgb
    )
    y, c1, c2 = from_xyz(xyz, video_parameters)

    return {
        "Y": y,
        "C1": c1,
        "C2": c2,
        "pic_num": picture_number,
    }


def enumerate_directories(dirname_a, dirname_b):
    """
    Given two directory names, return a list [(file_a, file_b), ...] of
    pictures to compare.

    Produces an error on stderr and call sys.exit if the files in the
    directories contain duplicate numbers, mismatched numbers or no numbered
    files at all.
    """
    file_lists = []  # [{number: filename_a, ...}, {number: filename_b, ...}]
    for dirname in [dirname_a, dirname_b]:
        # Find just the .raw files
        filenames = [
            os.path.join(dirname, name)
            for name in os.listdir(dirname)
            if name.endswith(".raw") and os.path.isfile(os.path.join(dirname, name))
        ]

        # Extract the numbers (checking for duplicates)
        numbered_filenames = {}  # {number: filename, ...}
        for filename in filenames:
            match = re.search(r"([0-9]+)\.raw$", filename)
            if match is None:
                sys.stderr.write(
                    "Error: An unnumbered file was found: {}\n".format(filename)
                )
                sys.exit(105)
            number = int(match.group(1))
            if number in numbered_filenames:
                sys.stderr.write(
                    "Error: More than one filename with number {} found in {}\n".format(
                        number,
                        dirname,
                    )
                )
                sys.exit(106)
            numbered_filenames[number] = filename

        if len(numbered_filenames) == 0:
            sys.stderr.write(
                "Error: No pictures found in directory {}\n".format(dirname)
            )
            sys.exit(107)

        file_lists.append(numbered_filenames)

    # Check numbers in filenames match
    filenames_a, filenames_b = file_lists
    unmatched_numbers = set(filenames_a).symmetric_difference(set(filenames_b))
    if unmatched_numbers:
        sys.stderr.write(
            "Error: Pictures numbered {} found in only one directory\n".format(
                ", ".join(map(str, sorted(unmatched_numbers)))
            )
        )
        sys.exit(108)

    return [
        (filename_a, filenames_b[num])
        for num, filename_a in sorted(filenames_a.items())
    ]


def compare_pictures(filename_a, filename_b, difference_mask_filename=None):
    """
    Compare a pair of pictures.

    Returns a string describing the differences
    between the pictures (if any) and integer return code which is non-zero iff
    the pictures differ.

    If a difference_mask filename is given, writes a difference mask to that
    file.
    """
    out = ""

    (
        (picture_a, video_parameters_a, picture_coding_mode_a),
        (picture_b, video_parameters_b, picture_coding_mode_b),
        byte_for_byte_identical,
    ) = read_pictures_with_only_one_metadata_file_required(filename_a, filename_b)

    if video_parameters_a != video_parameters_b:
        out += "Video parameters are different:\n"
        out += indent(video_parameter_diff(video_parameters_a, video_parameters_b))
        out += "\n"
        return (out.rstrip(), 1)

    if picture_coding_mode_a != picture_coding_mode_b:
        out += "Picture coding modes are different:\n"
        out += indent(
            picture_coding_mode_diff(picture_coding_mode_a, picture_coding_mode_b)
        )
        out += "\n"
        return (out.rstrip(), 2)

    if picture_a["pic_num"] != picture_b["pic_num"]:
        out += "Picture numbers are different:\n"
        out += indent(picture_number_diff(picture_a["pic_num"], picture_b["pic_num"]))
        out += "\n"
        return (out.rstrip(), 3)

    deltas = OrderedDict(
        (
            c,
            # NB: Convert pictures into Numpy arrays (for easier comparison)
            # using object mode (to ensure unlimited integer precision)
            np.array(picture_b[c], dtype=object) - np.array(picture_a[c], dtype=object),
        )
        for c in ["Y", "C1", "C2"]
    )

    identical, differences = measure_differences(
        deltas, video_parameters_a, picture_coding_mode_a
    )
    if identical:
        if not byte_for_byte_identical:
            out += (
                "Warning: Padding bits in raw picture data are different "
                "(is file endianness correct?)\n"
            )
        out += "Pictures are identical\n"
    else:
        out += "Pictures are different:\n"
        out += indent(differences)
        out += "\n"

    # Write difference mask, as required
    if difference_mask_filename is not None:
        mask = generate_difference_mask_picture(
            deltas,
            video_parameters_a,
            picture_number=picture_a["pic_num"],
        )
        write(mask, video_parameters_a, picture_coding_mode_a, difference_mask_filename)

    return (out.rstrip(), 0 if identical else 4)


def main(*args, **kwargs):
    args = parse_args(*args, **kwargs)

    a_is_dir = os.path.isdir(args.filename_a)
    b_is_dir = os.path.isdir(args.filename_b)

    if a_is_dir != b_is_dir:
        sys.stderr.write(
            "Error: Arguments must both be files or both be directories.\n"
        )
        return 103

    if a_is_dir and b_is_dir:
        if args.difference_mask is not None:
            sys.stderr.write(
                "Error: --difference-mask/-D can only be used "
                "when files (not directories) are compared.\n"
            )
            return 104

        final_exitcode = 0
        num_different = 0
        num_same = 0
        for filename_a, filename_b in enumerate_directories(
            args.filename_a, args.filename_b
        ):
            print("Comparing {} and {}:".format(filename_a, filename_b))
            message, exitcode = compare_pictures(filename_a, filename_b)
            print(indent(message))
            if exitcode != 0:
                final_exitcode = exitcode
                num_different += 1
            else:
                num_same += 1

        print("Summary: {} identical, {} different".format(num_same, num_different))
        return final_exitcode
    else:
        message, exitcode = compare_pictures(
            args.filename_a,
            args.filename_b,
            args.difference_mask,
        )
        print(message)
        return exitcode


if __name__ == "__main__":
    sys.exit(main())



"""
VC-2 style 2D array functions (5.5).
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode


__all__ = [
    "new_array",
    "width",
    "height",
    "row",
    "column",
    "delete_rows_after",
    "delete_columns_after",
]


@ref_pseudocode(deviation="inferred_implementation")
def new_array(*dimensions):
    """
    (5.4.2) Makes an N-dimensional array out of nested lists. Dimensions are
    given in the same order as they are indexed, e.g. ``new_array(height,
    width)``, like ``array[y][x]``.
    """
    if len(dimensions) == 0:
        return None
    else:
        return [new_array(*dimensions[1:]) for _ in range(dimensions[0])]


@ref_pseudocode(deviation="inferred_implementation")
def width(a):
    """(5.5.4)"""
    # NB: In zero-height arrays this will give the wrong answer (0), but such
    # arrays are not expected in reality.
    if len(a) == 0:
        return 0
    else:
        return len(a[0])


@ref_pseudocode(deviation="inferred_implementation")
def height(a):
    """(5.5.4)"""
    return len(a)


@ref_pseudocode(deviation="inferred_implementation")
def row(a, k):
    """
    (15.4.1) A 1D-array-like view into a row of a (2D) nested list as returned
    by :py:func:`new_array()`.
    """
    return a[k]


@ref_pseudocode(deviation="inferred_implementation")
class column(object):
    """
    (15.4.1) A 1D-array-like view into a column of a (2D) nested list as
    returned by :py:func:`new_array()`.
    """

    def __init__(self, a, k):
        self._a = a
        self._k = k

    def __getitem__(self, key):
        assert isinstance(key, int)
        return self._a[key][self._k]

    def __setitem__(self, key, value):
        assert isinstance(key, int)
        self._a[key][self._k] = value

    def __len__(self):
        return len(self._a)


def delete_rows_after(a, k):
    """
    (15.4.5) Delete rows 'k' and after in 'a'.
    """
    del a[k:]


def delete_columns_after(a, k):
    """
    (15.4.5) Delete columns 'k' and after in 'a'.
    """
    for row in a:
        del row[k:]



r"""
The :py:mod:`vc2_conformance.string_formatters` module contains facilities for
formatting (or pretty printing) Python values as strings.

When we say 'string formatter' we mean a function/callable which takes a value
and returns a string representation of that value. Instances of the classes in
this module act as formatters. For example, the :py:class:`Hex` class may be
used as a formatter for n-digit hexadecimal integers::

    >>> from vc2_conformance.string_formatters import Hex

    >>> # Create a formatter for producing 8-digit hex numbers
    >>> hex32_formatter = Hex(8)

    >>> # Format some values
    >>> hex32_formatter(0)
    '0x00000000'
    >>> hex32_formatter(0x1234)
    '0x00001234'

"""

from vc2_conformance.string_utils import indent, ellipsise

__all__ = [
    "Number",
    "Hex",
    "Dec",
    "Oct",
    "Bin",
    "Bool",
    "Bits",
    "Bytes",
    "Object",
    "List",
    "MultilineList",
]


class Number(object):
    """
    A formatter which uses Python's built-in :py:meth:`str.format` method to
    apply formatting.

    This formatter is quite low level, see :py:class:`Hex`, :py:class:`Dec`,
    :py:class:`Oct` and :py:class:`Bin` for ready to use derivatives.

    Parameters
    ==========
    format_code : str
        A python :py:meth:`str.format` code, e.g. "b" for binary.
    prefix : str
        A prefix to add before the formatted number
    num_digits : int
        The length to pad the number to.
    pad_digit : str
        The value to use to pad absent digits
    """

    def __init__(self, format_code, num_digits=0, pad_digit="0", prefix=""):
        self.format_code = format_code
        self.num_digits = num_digits
        self.pad_digit = pad_digit
        self.prefix = prefix

    def __call__(self, number):
        return "{}{}{:{}{}{}}".format(
            "-" if number < 0 else "",
            self.prefix,
            abs(number),
            self.pad_digit,
            self.num_digits,
            self.format_code,
        )


class Hex(Number):
    """
    Prints numbers in hexadecimal.

    Parameters
    ==========
    num_digits : int
        Minimum number of digits to show.
    pad_digit : str
        The value to use to pad absent digits
    prefix : str
        Defaults to "0x"
    """

    def __init__(self, num_digits=0, pad_digit="0", prefix="0x"):
        super(Hex, self).__init__("X", num_digits, pad_digit, prefix)


class Dec(Number):
    """
    Prints numbers in decimal.

    Parameters
    ==========
    num_digits : int
        Minimum number of digits to show.
    pad_digit : str
        The value to use to pad absent digits
    prefix : str
        Defaults to ""
    """

    def __init__(self, num_digits=0, pad_digit="0", prefix=""):
        super(Dec, self).__init__("d", num_digits, pad_digit, prefix)


class Oct(Number):
    """
    Prints numbers in octal.

    Parameters
    ==========
    num_digits : int
        Minimum number of digits to show.
    pad_digit : str
        The value to use to pad absent digits
    prefix : str
        Defaults to "0o"
    """

    def __init__(self, num_digits=0, pad_digit="0", prefix="0o"):
        super(Oct, self).__init__("o", num_digits, pad_digit, prefix)


class Bin(Number):
    """
    Prints numbers in binary.

    Parameters
    ==========
    num_digits : int
        Minimum number of digits to show.
    pad_digit : str
        The value to use to pad absent digits
    prefix : str
        Defaults to "0b"
    """

    def __init__(self, num_digits=0, pad_digit="0", prefix="0b"):
        super(Bin, self).__init__("b", num_digits, pad_digit, prefix)


class Bool(object):
    """
    A formatter for :py:class:`bool` (or bool-castable) objects. For the values
    0, 1, False and True, just shows 'True' or 'False'. For all other values,
    shows also the true value in brackets.

    For example::

        >>> bool_formatter = Bool()
        >>> bool_formatter(False)
        "False"
        >>> bool_formatter(True)
        "True"
        >>> bool_formatter(0)
        "False"
        >>> bool_formatter(1)
        "True"
        >>> bool_formatter(123)
        "True (123)"
        >>> bool_formatter(None)
        "True (None)"
    """

    def __call__(self, b):
        if b and b in (1, True):
            return "True"
        elif not b and b in (0, False):
            return "False"
        else:
            return "{} ({})".format(bool(b), b)


class Bits(object):
    """
    A formatter for :py:class:`bitarray.bitarray` objects. Shows the value as a
    string of the form '0b0101', using
    :py:func:`~vc2_conformance.string_utils.ellipsise` to shorten very long,
    values.

    Parameters
    ==========
    prefix : str
        A prefix to add to the string
    context : int
    min_length : int
        See :py:func:`~vc2_conformance.string_utils.ellipsise`.
    show_length : int or bool
        If an integer, show the length of the bitarray in brackets if above the
        specified length (in bits). If a bool, force display (or hiding) of the
        length information.
    """

    def __init__(self, prefix="0b", context=4, min_length=8, show_length=16):
        self.prefix = prefix
        self.context = context
        self.min_length = min_length
        self.show_length = show_length

    def __call__(self, ba):
        string = ellipsise(ba.to01(), self.context, self.min_length)
        if self.show_length is not False:
            if self.show_length is True or len(ba) >= self.show_length:
                string += " ({} bit{})".format(
                    len(ba),
                    "s" if len(ba) != 1 else "",
                )
        return "{}{}".format(self.prefix, string)


class Bytes(object):
    """
    A formatter for :py:class:`bytes` strings. Shows the value as a string of
    the form '0xAB_CD_EF', using
    :py:func:`~vc2_conformance.string_utils.ellipsise` to shorten very long,
    values.

    Parameters
    ==========
    prefix : str
        A prefix to add to the string
    separator : str
        A string to place between each pair of hex digits.
    context : int
    min_length : int
        See :py:func:`~vc2_conformance.string_utils.ellipsise`.
    show_length : int or bool
        If an integer, show the length of the bitarray in brackets if above the
        specified length (in bytes). If a bool, force display (or hiding) of
        the length information.
    """

    def __init__(
        self, prefix="0x", separator="_", context=2, min_length=4, show_length=8
    ):
        self.prefix = prefix
        self.separator = separator
        self.context = context
        self.min_length = min_length
        self.show_length = show_length

    def __call__(self, b):
        string = "".join("{:02X}".format(n) for n in bytearray(b))

        before, ellipses, after = ellipsise(
            string, self.context * 2, self.min_length * 2
        ).partition("...")

        # Interleave bytes with separator
        before = "".join(
            c
            if (i % 2 == 0 or i == len(before) - 1)
            else "{}{}".format(c, self.separator)
            for i, c in enumerate(before)
        )
        after = "".join(
            c
            if ((len(after) - i) % 2 == 1 or i == 0)
            else "{}{}".format(self.separator, c)
            for i, c in enumerate(after)
        )

        string = before + ellipses + after

        if self.show_length is not False:
            if self.show_length is True or len(b) >= self.show_length:
                string += " ({} byte{})".format(
                    len(b),
                    "s" if len(b) != 1 else "",
                )
        return "{}{}".format(self.prefix, string)


class Object(object):
    """
    A formatter for opaque python Objects. Shows only the object type name.
    """

    def __init__(self, prefix="<", suffix=">"):
        self.prefix = prefix
        self.suffix = suffix

    def __call__(self, o):
        return "{}{}{}".format(
            self.prefix,
            type(o).__name__,
            self.suffix,
        )


class List(object):
    """
    A formatter for lists which collapses repeated entries.

    Examples::

        >>> # Use Python-style notation for repeated entries
        >>> List()([1, 1, 1, 1])
        [1]*4

        >>> # Also displays lists with some non-repeated values
        >>> List()([1, 2, 3, 0, 0, 0, 0, 0, 4, 5])
        [1, 2, 3] + [0]*5 + [4, 5]

        >>> # A custom formatter may be supplied for formatting the list
        >>> # entries
        >>> List(formatter=Hex())([1, 2, 3, 0, 0, 0])
        [0x1, 0x2, 0x3] + [0x0]*3

        >>> # Equality is based on the string formatted value, not the raw
        >>> # value
        >>> List(formatter=Object())([1, 2, 3, 0, 0, 0])
        [<int>]*6

        >>> # The minimum run-length before truncation may be overridden
        >>> List(min_run_length=3)([1, 2, 2, 3, 3, 3])
        [1, 2, 2] + [3]*3
    """

    def __init__(self, min_run_length=3, formatter=str):
        self.min_run_length = min_run_length
        self.formatter = formatter

    def __call__(self, lst):
        # Special case (avoids complications below)
        if len(lst) == 0:
            return "[]"

        values = [self.formatter(v) for v in lst]

        # For each value in the input list, count the run length at that point
        # of that value.
        run_lengths = [1]
        for last_value, value in zip(values, values[1:]):
            if last_value == value:
                run_lengths.append(run_lengths[-1] + 1)
            else:
                run_lengths.append(1)

        # Accumulate a list of lists and tuples. For every series of
        # non-identical values, a list of values will be included. For every
        # run of values, a (value, run_length) tuple will be included.
        out = [[]]
        while values:
            run_length = run_lengths.pop()
            value = values.pop()

            if run_length < self.min_run_length:
                out[0].insert(0, value)
            else:
                if len(out[0]) == 0:
                    del out[0]
                out.insert(0, (value, run_length))
                out.insert(0, [])

                del run_lengths[-(run_length - 1) :]
                del values[-(run_length - 1) :]

        if len(out[0]) == 0:
            del out[0]

        # Format as a string
        return " + ".join(
            (
                "[{}]".format(", ".join(value_run))
                if isinstance(value_run, list)
                else "[{}]*{}".format(*value_run)
            )
            for value_run in out
        )


class MultilineList(object):
    """
    A formatter for lists which displays each value on its own line.

    Examples::

        >>> MultilineList()(["one", "two", "three"])
        0: one
        1: two
        2: three

        >>> # A custom formatter may be supplied for formatting the list
        >>> # entries
        >>> MultilineList(formatter=Hex())([1, 2, 3])
        0: 0x1
        1: 0x2
        2: 0x3

        >>> # A heading may be added
        >>> MultilineList(heading="MyList")(["one", "two", "three"])
        MyList
          0: one
          1: two
          2: three
    """

    def __init__(self, heading=None, formatter=str):
        self.heading = heading
        self.formatter = formatter

    def __call__(self, lst):
        lines = "\n".join(
            "{}: {}".format(i, self.formatter(value)) for i, value in enumerate(lst)
        )

        if self.heading is None:
            return lines
        else:
            return "{}\n{}".format(self.heading, indent(lines))



"""
Real picture based test sequences.
"""

from vc2_conformance.bitstream import Stream

from vc2_conformance.encoder import make_sequence

from vc2_conformance import picture_generators

from vc2_conformance.test_cases import decoder_test_case_generator


@decoder_test_case_generator
def real_pictures(codec_features):
    """
    **Tests real pictures are decoded correctly.**

    A series of three still photographs.

    .. image:: /_static/user_guide/real_pictures.svg

    .. note::

        The images encoded in this sequence are generated from 4256 by 2832
        pixel, 4:4:4, 16 bit, standard dynamic range, RGB color images with the
        ITU-R BT.709 gamut. As such, the decoded pictures might be of reduced
        technical quality compared with the capabilities of the format. The
        rescaling, color conversion and encoding algorithms used are also basic
        in nature, potentially further reducing the picture quality.
    """
    return Stream(
        sequences=[
            make_sequence(
                codec_features,
                picture_generators.real_pictures(
                    codec_features["video_parameters"],
                    codec_features["picture_coding_mode"],
                ),
            )
        ]
    )



"""
Test pictures designed to produce extreme signal levels in an encoder.
"""

import logging

from vc2_conformance.test_cases import (
    encoder_test_case_generator,
    TestCase,
    EncoderTestSequence,
)

from vc2_conformance.picture_generators import (
    mid_gray,
    repeat_pictures,
)

from vc2_conformance.test_cases.bit_widths_common import (
    get_test_pictures,
    MissingStaticAnalysisError,
)


@encoder_test_case_generator
def signal_range(codec_features):
    """
    **Tests that an encoder has sufficient numerical dynamic range.**

    These test cases contain test patterns designed to produce extreme signals
    within encoders. During these test cases, no integer clamping or overflows
    must occur.

    A test case is produced for each picture component:

    ``signal_range[Y]``
        Luma component test patterns.

    ``signal_range[C1]``
        Color difference 1 component test patterns.

    ``signal_range[C2]``
        Color difference 2 component test patterns.

    Though the test patterns produce near worst case signal levels, they are
    not guaranteed to produce the largest values possible.

    .. note::

        For informational purposes, an example of a set of test patterns are shown below:

        .. image:: /_static/user_guide/signal_range_encoder.png

    An informative metadata file is provided along side each test case which
    gives, for each picture in the bitstream, the parts of a encoder which are
    being tested by the test patterns. See
    :py:class:`vc2_bit_widths.helpers.TestPoint` for details.
    """
    try:
        (
            analysis_luma_pictures,
            synthesis_luma_pictures,
            analysis_color_diff_pictures,
            synthesis_color_diff_pictures,
        ) = get_test_pictures(codec_features)
    except MissingStaticAnalysisError:
        logging.warning(
            (
                "No static analysis available for the wavelet "
                "used by codec '%s'. Signal range test cases cannot "
                "be generated."
            ),
            codec_features["name"],
        )
        return

    for component, analysis_test_pictures in [
        ("Y", analysis_luma_pictures),
        ("C1", analysis_color_diff_pictures),
        ("C2", analysis_color_diff_pictures),
    ]:
        # Generate an initially empty set of mid-gray pictures
        one_gray_frame = list(
            mid_gray(
                codec_features["video_parameters"],
                codec_features["picture_coding_mode"],
            )
        )
        pictures = list(
            repeat_pictures(
                one_gray_frame,
                (
                    (len(analysis_test_pictures) + len(one_gray_frame) - 1)
                    // len(one_gray_frame)
                ),
            )
        )

        # Fill-in the test patterns
        for test_picture, picture in zip(analysis_test_pictures, pictures):
            picture[component] = test_picture.picture.tolist()

        # Extract the testpoints in JSON-serialisable form
        metadata = [
            [tp._asdict() for tp in p.test_points] for p in analysis_test_pictures
        ]

        out = EncoderTestSequence(
            pictures=pictures,
            video_parameters=codec_features["video_parameters"],
            picture_coding_mode=codec_features["picture_coding_mode"],
        )

        yield TestCase(out, component, metadata=metadata)



"""
The :py:mod:`vc2_conformance.codec_features` module defines the
:py:class:`CodecFeatures` :py:mod:`~vc2_conformance.fixeddict` which is used to
describe codec and video format configurations. These are used to control the
test case generators (:py:mod:`vc2_conformance.test_cases`) and encoder
(:py:mod:`vc2_conformance.encoder`).

.. autofixeddict:: CodecFeatures

The :py:func:`read_codec_features_csv` may be used to read these structures
from a CSV. This functionality is used by the :ref:`vc2-test-case-generator`
script.

.. autofunction:: read_codec_features_csv

.. autoexception:: InvalidCodecFeaturesError

Finally, the following function may be used when determining level-related
restrictions which apply to a set of :py:class:`CodecFeatures`.

.. autofunction:: codec_features_to_trivial_level_constraints

"""

import csv

from itertools import count, islice, product

from string import ascii_uppercase

from functools import partial

from fractions import Fraction

from collections import OrderedDict

from vc2_data_tables import (
    Levels,
    Profiles,
    PictureCodingModes,
    WaveletFilters,
    ColorDifferenceSamplingFormats,
    BaseVideoFormats,
    SourceSamplingModes,
    PresetColorPrimaries,
    PresetColorMatrices,
    PresetTransferFunctions,
)

from vc2_conformance.pseudocode.video_parameters import (
    set_source_defaults,
    picture_dimensions,
)

from vc2_conformance.pseudocode.state import State

from vc2_conformance.pseudocode.slice_sizes import slices_have_same_dimensions

from vc2_conformance.fixeddict import fixeddict, Entry


__all__ = [
    "CodecFeatures",
    "InvalidCodecFeaturesError",
    "read_codec_features_csv",
    "codec_features_to_trivial_level_constraints",
]


CodecFeatures = fixeddict(
    "CodecFeatures",
    Entry("name"),
    # (11.2.1)
    Entry("level", enum=Levels),
    Entry("profile", enum=Profiles),
    # (11.1)
    Entry("picture_coding_mode", enum=PictureCodingModes),
    # (11.4)
    Entry("video_parameters"),  # VideoParameters
    # (12.4.1) and (12.4.4.1)
    Entry("wavelet_index", enum=WaveletFilters),
    Entry("wavelet_index_ho", enum=WaveletFilters),
    Entry("dwt_depth"),
    Entry("dwt_depth_ho"),
    # (12.4.5.2)
    Entry("slices_x"),
    Entry("slices_y"),
    # (14)
    Entry("fragment_slice_count"),
    # Bitrate control
    Entry("lossless"),  # Bool
    Entry("picture_bytes"),  # None if lossless, int otherwise
    # (12.4.5.3)
    Entry("quantization_matrix"),
)
"""
A definition of a set of coding features supported by a video codec
implementation. In practice, a particular codec's feature set may be defined by
a collection of these, defining support for several picture formats.

Parameters
==========
name : str
    A unique, human-readable name to identify this collection of features (e.g.
    ``"uhd_over_hd_sdi"``).
level : :py:class:`~vc2_data_tables.Levels`
    The VC-2 level. The user is responsible for choosing settings below which
    actually conform to this level.
profile : :py:class:`~vc2_data_tables.Profiles`
    The VC-2 profile to use.
picture_coding_mode : :py:class:`~vc2_data_tables.PictureCodingModes`
    The picture coding mode to use.
video_parameters : :py:class:`~vc2_conformance.pseudocode.video_parameters.VideoParameters`
    The video format to use.
wavelet_index, wavelet_index_ho : :py:class:`~vc2_data_tables.WaveletFilters`
    The wavelet transform to use. For symmetric transforms, both values must be
    equal.
dwt_depth and dwt_depth_ho : int
    The transform depths to use. For symmetric transforms, dwt_depth_ho must be
    zero.
slices_x and slices_y : int
    The number of picture slices, horizontally and vertically.
fragment_slice_count : int
    If fragmented pictures are in use, should be non-zero and contain the
    maximum number of slices to include in each fragment. Otherwise, should be
    zero.
lossless : bool
    If True, lossless variable-bit-rate coding will be used. If false,
    fixed-rate lossy coding is used.
picture_bytes : int or None
    When ``lossless`` is False, this gives the number of bytes per picture to
    use. Slices will be assigned (as close to) the same number of bytes each as
    possible. If ``lossless`` is True, this value should be None.
quantization_matrix : None or {level: {orient: value, ...}, ...}
    None or a hierarchy of dictionaries as constructed by the ``quant_matrix``
    pseudocode function (12.4.5.3). If None, the default quantization matrix
    will be used.
"""


def read_dict_list_csv(csvfile):
    """
    Read a CSV whose first column contains key names and each subsequent column
    describes a dictionary with corresponding values.

    For example::

        a,one,ayy
        b,two,bee
        c,three,see

    Would become::

        [
            {
                "a": "one",
                "b": "two",
                "c": "three",
            },
            {
                "a": "ayy",
                "b": "bee",
                "c": "see",
            },
        ]

    The first column always gives dictionary keys. Rows whose first column is
    empty or contains only a string starting with a ``#`` (hash) character
    are ignored entirely. Leading and trailing whitespace will be stripped.
    """
    out = []
    for row in csv.reader(csvfile):
        # Skip empty rows
        if len(row) == 0:
            continue

        # Skip comment rows or rows with an empty first cell
        key = row[0].strip()
        if not key or key.startswith("#"):
            continue

        # Accumulate values
        for i, value in enumerate(row[1:]):
            if i >= len(out):
                out.append({})
            value = value.strip()
            if value:
                out[i][key] = value.strip()

    return out


def spreadsheet_column_names():
    """
    Generates an infinite sequence of spreadsheet column names (e.g. A, B, ...,
    Y, Z, AA, AB, ...).
    """
    for length in count(1):
        for name in product(ascii_uppercase, repeat=length):
            yield "".join(name)


def parse_int_enum(int_enum_type, value):
    """
    Parse a string containing an :py:class:`~enum.IntEnum` name or integer
    literal into the corresponding type.
    """
    try:
        number = int(value)
    except ValueError:
        for entry in int_enum_type:
            if entry.name == value:
                number = entry
                break
        else:
            raise ValueError(
                "{} is not a valid {}".format(
                    value,
                    int_enum_type.__name__,
                )
            )

    return int_enum_type(number)


def parse_int_at_least(minimum, value):
    """
    Parse a string containing an integer, raising a :py:exc:`ValueError` if the
    value is below the specified threshold.
    """
    value = int(value)
    if value < minimum:
        raise ValueError("{} < {}".format(value, minimum))
    else:
        return value


def parse_bool(value):
    """
    Parse a wide-ish array of CSV-style bool values.
    """
    lower_value = value.lower()

    if lower_value in ("1", "true", "t", "y", "yes"):
        return True
    elif lower_value in ("0", "false", "f", "n", "no"):
        return False
    else:
        raise ValueError("{} is not a valid boolean".format(value))


def parse_quantization_matrix(dwt_depth, dwt_depth_ho, value):
    """
    Parse a quantization matrix, presented as a series of whitespace-separated
    integers. Values are given in ascending order of level and subands are
    ordered as L, LL, H, HL, LH, HH (i.e. bitstream order in (12.4.5.3)).
    """
    values = iter(filter(None, value.split()))

    out = {}

    try:
        if dwt_depth_ho == 0:
            out[0] = {"LL": int(next(values))}
        else:
            out[0] = {"L": int(next(values))}
            for level in range(1, dwt_depth_ho + 1):
                out[level] = {"H": int(next(values))}
        for level in range(dwt_depth_ho + 1, dwt_depth + dwt_depth_ho + 1):
            out[level] = {
                "HL": int(next(values)),
                "LH": int(next(values)),
                "HH": int(next(values)),
            }
    except StopIteration:
        raise ValueError(
            "Expected {} values in quantisation matrix.".format(
                dwt_depth + dwt_depth_ho + 1
            )
        )

    try:
        next(values)
        raise ValueError(
            "Expected {} values in quantisation matrix.".format(
                dwt_depth + dwt_depth_ho + 1
            )
        )
    except StopIteration:
        pass

    return out


class InvalidCodecFeaturesError(ValueError):
    """
    Raised by :py:func:`read_codec_features_csv` encounters a problem with the
    data presented in a codec features listing CSV file.
    """


def read_codec_features_csv(csvfile):
    """
    Read a set of :py:class:`CodecFeatures` from a CSV file in the format
    described by :ref:`codec-features`.

    ..
        If you're reading this, you probably can't click on the reference
        above... It points to the 'Defining codec features' section of
        ``docs/user_guide/generating_test_cases.rst``

    Parameters
    ==========
    csvfile : iterable
        An iterable of lines from a CSV file, e.g. an open :py:class:`file`
        object.

    Returns
    =======
    codec_feature_sets : OrderedDict([(name, :py:class:`CodecFeatures`), ...])

    Raises
    ======
    :py:exc:`InvalidCodecFeaturesError`
        Raised if the provided CSV contains invalid or incomplete data.

        .. note::

            Validation largely extends only to syntactic issues (e.g. invalid
            integer values, 'picture_bytes' being specified for lossless
            formats etc). It does not include validation of 'deeper' issues
            such as too-small picture_bytes values or parameters not being
            permitted by the specified level.
    """
    csv_columns = read_dict_list_csv(csvfile)

    out = OrderedDict()

    for i, column in zip(islice(spreadsheet_column_names(), 1, None), csv_columns):
        if not column:
            continue

        def pop(field_name, parser, *default_):
            """
            Check for the existance of a value in the current column and return
            the native-type version of it. If a third argument, 'default', is
            given, the default value will be returned if the field contains the
            string "default".
            """
            assert len(default_) in (0, 1)
            try:
                value = column.pop(field_name)
                if default_ and value.lower() == "default":
                    return default_[0]
                else:
                    return parser(value)
            except KeyError:
                raise InvalidCodecFeaturesError(
                    "Missing entry for '{}' in '{}' column".format(
                        field_name,
                        name,
                    )
                )
            except ValueError as e:
                raise InvalidCodecFeaturesError(
                    "Invalid entry for '{}' in '{}' column: {} ({})".format(
                        field_name,
                        name,
                        value,
                        e,
                    )
                )

        features = CodecFeatures()

        # Create default names for columns where not provided
        if "name" not in column:
            name = "column_{}".format(i)
        else:
            name = column.pop("name")
        name = name.strip()

        features["name"] = name

        # Check for name uniqueness
        if name in out:
            raise InvalidCodecFeaturesError(
                "Name '{}' used more than once".format(name)
            )

        out[name] = features

        # Parse basic fields
        for field_name, field_type in [
            ("level", partial(parse_int_enum, Levels)),
            ("profile", partial(parse_int_enum, Profiles)),
            ("picture_coding_mode", partial(parse_int_enum, PictureCodingModes)),
            ("wavelet_index", partial(parse_int_enum, WaveletFilters)),
            ("wavelet_index_ho", partial(parse_int_enum, WaveletFilters)),
            ("dwt_depth", partial(parse_int_at_least, 0)),
            ("dwt_depth_ho", partial(parse_int_at_least, 0)),
            ("slices_x", partial(parse_int_at_least, 1)),
            ("slices_y", partial(parse_int_at_least, 1)),
            ("fragment_slice_count", partial(parse_int_at_least, 0)),
            ("lossless", parse_bool),
        ]:
            features[field_name] = pop(field_name, field_type)

        features["video_parameters"] = set_source_defaults(
            pop(
                "base_video_format",
                partial(parse_int_enum, BaseVideoFormats),
            )
        )

        # Parse integer video_parameters fields
        for field_name, field_type in [
            ("frame_width", partial(parse_int_at_least, 1)),
            ("frame_height", partial(parse_int_at_least, 1)),
            (
                "color_diff_format_index",
                partial(parse_int_enum, ColorDifferenceSamplingFormats),
            ),
            ("source_sampling", partial(parse_int_enum, SourceSamplingModes)),
            ("top_field_first", parse_bool),
            ("frame_rate_numer", partial(parse_int_at_least, 1)),
            ("frame_rate_denom", partial(parse_int_at_least, 1)),
            ("pixel_aspect_ratio_numer", partial(parse_int_at_least, 1)),
            ("pixel_aspect_ratio_denom", partial(parse_int_at_least, 1)),
            ("clean_width", partial(parse_int_at_least, 0)),
            ("clean_height", partial(parse_int_at_least, 0)),
            ("left_offset", partial(parse_int_at_least, 0)),
            ("top_offset", partial(parse_int_at_least, 0)),
            ("luma_offset", partial(parse_int_at_least, 0)),
            ("luma_excursion", partial(parse_int_at_least, 1)),
            ("color_diff_offset", partial(parse_int_at_least, 0)),
            ("color_diff_excursion", partial(parse_int_at_least, 1)),
            ("color_primaries_index", partial(parse_int_enum, PresetColorPrimaries)),
            ("color_matrix_index", partial(parse_int_enum, PresetColorMatrices)),
            (
                "transfer_function_index",
                partial(parse_int_enum, PresetTransferFunctions),
            ),
        ]:
            features["video_parameters"][field_name] = pop(
                field_name,
                field_type,
                features["video_parameters"][field_name],
            )

        # Parse picture_bytes option
        if features["lossless"]:
            if "picture_bytes" in column:
                raise InvalidCodecFeaturesError(
                    "Entry provided for 'picture_bytes' when lossless mode "
                    "specified for '{}' column".format(
                        name,
                    )
                )
            features["picture_bytes"] = None
        else:
            features["picture_bytes"] = pop(
                "picture_bytes",
                partial(parse_int_at_least, 1),
            )

        # Parse quantisation matrix
        features["quantization_matrix"] = pop(
            "quantization_matrix",
            partial(
                parse_quantization_matrix,
                features["dwt_depth"],
                features["dwt_depth_ho"],
            ),
            None,
        )

        # Check for extraneous rows
        if column:
            raise InvalidCodecFeaturesError(
                "Unrecognised row(s): {}".format(
                    ", ".join(set(column)),
                )
            )

    return out


def codec_features_to_trivial_level_constraints(codec_features):
    """
    Returns the some of the values a given :py:class:`CodecFeatures` would
    assign in a :py:mod:`~vc2_conformance.level_constraints` table.

    Parameters
    ==========
    codec_features : :py:class:`CodecFeatures`

    Returns
    =======
    constrained_values : {key: concrete_value, ...}
        A partial set of :py:mod:`~vc2_conformance.level_constraints`,
        specifically containing the following keys:

        * level
        * profile
        * picture_coding_mode
        * wavelet_index
        * dwt_depth
        * slices_x
        * slices_y
        * slices_have_same_dimensions
        * custom_quant_matrix

        In addition for the low delay profile only:

        * slice_bytes_numerator
        * slice_bytes_denominator

        In addition for the high quality profile only:

        * slice_prefix_bytes

        .. note::

            In principle, more keys could be determined, however a line in the
            sand is required for what is considered 'simple' to determine and
            what requires re-implementing much of the codec. We draw the line
            at these values since all of them are straight-forward to work out.
    """
    constrained_values = {}

    # Copy across the trivial options
    for key in [
        "level",
        "profile",
        "picture_coding_mode",
        "wavelet_index",
        "dwt_depth",
        "slices_x",
        "slices_y",
    ]:
        constrained_values[key] = codec_features[key]

    # Determine if all slices have the same dimensions
    state = State(
        dwt_depth=codec_features["dwt_depth"],
        dwt_depth_ho=codec_features["dwt_depth_ho"],
        slices_x=codec_features["slices_x"],
        slices_y=codec_features["slices_y"],
        picture_coding_mode=codec_features["picture_coding_mode"],
    )
    picture_dimensions(
        state,
        codec_features["video_parameters"],
    )
    constrained_values["slices_have_same_dimensions"] = slices_have_same_dimensions(
        state
    )

    # Determine slice sizes
    if codec_features["profile"] == Profiles.low_delay:
        num_slices = codec_features["slices_x"] * codec_features["slices_y"]
        slice_bytes = Fraction(codec_features["picture_bytes"] or 0, num_slices)
        constrained_values["slice_bytes_numerator"] = slice_bytes.numerator
        constrained_values["slice_bytes_denominator"] = slice_bytes.denominator
    elif codec_features["profile"] == Profiles.high_quality:
        # NB: At the moment the test case generator cannot generate streams
        # which include prefix bytes so we assume it is zero below... (See
        # `vc2_conformance.encoder.pictures.make_picture_parse`)
        constrained_values["slice_prefix_bytes"] = 0

    # Determine custom quantisation matrix values
    constrained_values["custom_quant_matrix"] = (
        codec_features["quantization_matrix"] is not None
    )

    return constrained_values



"""
Picture component offsetting routines (15)

This module collects picture component value offsetting functions from (15) and
augments these with complementary de-offsetting functions (inferred from, but
not defined by the standard).
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.pseudocode.arrays import (
    width,
    height,
)


__all__ = [
    "offset_picture",
    "offset_component",
    "remove_offset_picture",
    "remove_offset_component",
]


@ref_pseudocode
def offset_picture(state, current_picture):
    """(15.5) Remove picture value offsets (added during encoding).

    Parameters
    ==========
    state : :py:class:`vc2_conformance.pseudocode.state.State`
        Where ``luma_depth`` and ``color_diff_depth`` are defined.
    current_picture : {comp: [[pixel_value, ...], ...], ...}
        Will be mutated in-place.
    """
    for c in ["Y", "C1", "C2"]:
        offset_component(state, current_picture[c], c)


@ref_pseudocode
def offset_component(state, comp_data, c):
    """(15.5) Remove picture value offsets from a single component."""
    for y in range(height(comp_data)):
        for x in range(width(comp_data)):
            if c == "Y":
                comp_data[y][x] += 2 ** (state["luma_depth"] - 1)
            else:
                comp_data[y][x] += 2 ** (state["color_diff_depth"] - 1)


def remove_offset_picture(state, current_picture):
    """Inverse of offset_picture (15.5). Centers picture values around zero."""
    for c in ["Y", "C1", "C2"]:
        remove_offset_component(state, current_picture[c], c)


def remove_offset_component(state, comp_data, c):
    """
    Inverse of offset_component (15.5). Centers picture values around zero.

    Parameters
    ==========
    state : :py:class:`vc2_conformance.pseudocode.state.State`
        Where ``luma_depth`` and ``color_diff_depth`` are defined.
    current_picture : {comp: [[pixel_value, ...], ...], ...}
        Will be mutated in-place.
    """
    for y in range(height(comp_data)):
        for x in range(width(comp_data)):
            if c == "Y":
                comp_data[y][x] -= 2 ** (state["luma_depth"] - 1)
            else:
                comp_data[y][x] -= 2 ** (state["color_diff_depth"] - 1)



"""
The :py:mod:`vc2_conformance.string_utils` module contains a selection of
general purpose string formatting routines.
"""

from textwrap import wrap, dedent

import re


def indent(text, prefix="  "):
    """
    Indent the string 'text' with the prefix string 'prefix'.

    .. note::

        This function is provided partly because Python 2.x doesn't include
        :py:func:`textwrap.indent` in its standard library and partly to
        provide an indent function with sensible defaults (i.e. 2 character
        indent, and always indent every line).
    """
    return "{}{}".format(prefix, ("\n{}".format(prefix)).join(text.split("\n")))


def ellipsise(text, context=4, min_length=8):
    """
    Given a string which contains very long sequences of the same character
    (e.g. long mostly constant binary or hex numbers), produce an 'ellipsised'
    version with some of the repeated characters replaced with '...'.

    Exactly one shortening operation will be carried out (on the longest run)
    meaning that so long as the original string length is known, no ambiguity
    is introduced in the ellipsised version.

    For example::

        >>> ellipsise("0b10100000000000000000000000000000000000001")
        "0b1010000...00001"

    Parameters
    ==========
    text : str
        String to ellipsise.
    context : int
        The number of repeated characters to retain before and after the
        ellipses.
    min_length : int
        The minimum number of characters to bother replacing with '...'. This
        means that no change will be made until 2*context + min_length
        character repetitions.
    """
    # Special case to avoid handling it later
    if len(text) == 0:
        return text

    repeats = []
    num_repeats = 0
    last_char = None
    for char in text:
        if char == last_char:
            num_repeats += 1
        else:
            num_repeats = 1
        repeats.append(num_repeats)
        last_char = char

    longest_run_end, longest_run_length = max(
        enumerate(repeats),
        key=lambda pos_count: pos_count[1],
    )

    if longest_run_length < (2 * context) + min_length:
        # Too short to bother
        return text
    else:
        longest_run_start = longest_run_end - longest_run_length + 1
        return "{}...{}".format(
            text[: longest_run_start + context],
            text[longest_run_end - context + 1 :],
        )


def ellipsise_lossy(text, max_length=80):
    """
    Given a string which may not fit within a given line length, trnucate the
    string by adding ellipses in the middle.
    """
    if len(text) <= max_length:
        # Already short enough
        return text
    else:
        before_length = (max_length - 3) // 2
        after_length = (max_length - 3) - before_length
        return "{}...{}".format(text[:before_length], text[-after_length:])


RE_HEADING_UNDERLINE = re.compile(r"^[-=]+$")
RE_BULLET_OR_NUMBER = re.compile(r"^([*]\s+|[0-9]+[.]\s+)(.*)$")
RE_PREFIX_AND_BLOCK = re.compile(r"^([*]\s+|[0-9]+[.]\s+|\s*)(.*)$")


def split_into_line_wrap_blocks(text, wrap_indented_blocks=False):
    """
    Deindent and split a multi-line markdown-style string into blocks of text
    which can be line-wrapped independently.

    For example given a markdown-style string defined like so::

        '''
            A markdown style title
            ======================

            This is a string with some initial indentation
            and also some hard line-wraps inserted too. This
            paragraph ought to be line-wrapped as an
            independent unit.

            Here's a second paragraph which also ought to be
            line wrapped as its own unit.

            * This is a bulleted list
            * Each bullet point should be line wrapped as an
              individual unit (with the wrapping indented
              as shown here).
            * Notice that bullets don't have a newline
              between them like paragraphs do.

            1. Numbered lists are also supported.
            2. Here long lines will be line wrapped in much
               the same way as a bulleted list.

            Finally:

                An intended block will also remain indented.
                However, if wrap_indented_blocks is False, the
                existing linebreaks will be retained (e.g. for
                markdown-style code blocks). If set to True,
                the indented block will be line-wrapped.
        '''

    This will be split into independently line wrappable segments (as
    described).

    Returns
    -------
    blocks : [(first_indent, rest_indent, text), ...]
        A series of wrappable blocks. In each tuple:

        * first_indent contains a string which should be used to indent the
          first line of the wrapped block.
        * rest_indent should be a string which should be used to indent all
          subsequent lines in the wrapped block. This will be the same length
          as first_indent.
        * text will be an indentation and newline-free string

        An empty block (i.e. ``("", "", "")``) will be included between
        each paragraph in the input so that the output maintains the same
        vertical whitespace profile.
    """
    # Remove common leading whitespace.
    text = dedent(text)

    block_lines = [""]
    for line in text.splitlines():
        # Start a new block if we encounter an empty line (between paragraphs)
        # or a bullet point/number. The first bullet/number must have a blank
        # line before it.
        if line.rstrip() == "":
            block_lines.append("")
            block_lines.append("")
        elif RE_HEADING_UNDERLINE.match(line):
            block_lines.append("")
        elif RE_BULLET_OR_NUMBER.match(line) and (
            # Don't match asterisks and number-dots midway through
            # paragraphs -- require that they appear after a blank line or
            # after another bullet/number.
            block_lines[-1] == ""
            or RE_BULLET_OR_NUMBER.match(block_lines[-1])
        ):
            block_lines.append("")
        elif not wrap_indented_blocks and re.match(r"^\s+", block_lines[-1]):
            # When block wrapping is disabled, start a new line for every line
            # in the block
            block_lines.append("")

        # Retain indentation of first non-empty line in block
        if block_lines[-1].strip() == "":
            block_lines[-1] = line.rstrip() + " "
        else:
            block_lines[-1] += line.strip() + " "

    # Remove leading and repeating blank lines
    index = 0
    previous_line_is_blank = True
    while index < len(block_lines):
        is_blank = block_lines[index].strip() == ""

        if is_blank and previous_line_is_blank:
            del block_lines[index]
        else:
            index += 1

        previous_line_is_blank = is_blank

    # Remove trailing blank lines
    while block_lines and block_lines[-1].strip() == "":
        del block_lines[-1]

    out = []
    for block_line in block_lines:
        prefix, text = RE_PREFIX_AND_BLOCK.match(block_line).groups((1, 2))
        if len(prefix.strip()) > 0:
            out.append((prefix, " " * len(prefix), text.rstrip()))
        else:
            out.append((prefix, prefix, text.rstrip()))

    return out


def wrap_blocks(blocks, width=None, wrap_indented_blocks=False):
    """
    Return a line-wrapped version of a series of text blocks as produced by
    :py:func:`split_into_line_wrap_blocks`.

    Expects a list of (first_line_indent, remaining_line_indent, text) tuples
    to output.

    If 'width' is None, assumes an infinite line width.

    If 'wrap_indented_blocks' is False (the default) indented (markdown-style)
    code blocks will not be line wrapped while other indented blocks (e.g.
    bullets) will be.
    """
    return "\n".join(
        first_indent
        + ("\n" + rest_indent).join(
            (
                wrap(text, width - len(first_indent))
                if (
                    wrap_indented_blocks
                    or len(first_indent) == 0
                    or first_indent != rest_indent
                )
                else [text]
            )
            if width is not None
            else [text]
        )
        for first_indent, rest_indent, text in blocks
    )


def wrap_paragraphs(text, width=None, wrap_indented_blocks=False):
    """
    Re-line-wrap a markdown-style string with hard-line-wrapped paragraphs,
    bullet points, numbered lists and code blocks (see
    :py:func:`split_into_line_wrap_blocks`).

    If 'width' is None, assumes an infinite line width.

    If 'wrap_indented_blocks' is False (the default) indented (markdown-style)
    code blocks will not be line wrapped while other indented blocks (e.g.
    bullets) will be.
    """
    return wrap_blocks(
        split_into_line_wrap_blocks(text, wrap_indented_blocks),
        width,
        wrap_indented_blocks,
    )



"""
Slice size computation functions (13)

All of the functions below take a 'state' argument which should be a dictionary-like
object containing the following entries:

* ``"luma_width"``
* ``"luma_height"``
* ``"color_diff_width"``
* ``"color_diff_height"``
* ``"dwt_depth"``
* ``"dwt_depth_ho"``

The :py:func:`slice_bytes`, :py:func:`slice_top`, :py:func:`slice_bottom`,
:py:func:`slice_left`, :py:func:`slice_right` and
:py:func:`slices_have_same_dimensions` functions all additionally require the
following entries:

* ``"slices_x"``
* ``"slices_y"``

Finally :py:func:`slice_bytes` function requires the following further values:

* ``"slice_bytes_numerator"``
* ``"slice_bytes_denominator"``

The 'c' or 'comp' arguments should be set to one of the following strings:

* ``"Y"``
* ``"C1"``
* ``"C2"``

The :py:func:`slices_have_same_dimensions` utility is added beyond the VC-2
pseudocode functions which determines if all slices will contain the same
number of samples or not.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

__all__ = [
    "subband_width",
    "subband_height",
    "slice_bytes",
    "slice_left",
    "slice_right",
    "slice_top",
    "slice_bottom",
    "slices_have_same_dimensions",
]


@ref_pseudocode(deviation="inferred_implementation")
def subband_width(state, level, comp):
    """(13.2.3)"""
    if comp == "Y":
        w = state["luma_width"]
    elif comp == "C1" or comp == "C2":
        w = state["color_diff_width"]

    # Round up (pad) the picture width to the nearest multiple of the scale
    # width
    scale_w = 1 << (state["dwt_depth_ho"] + state["dwt_depth"])
    pw = scale_w * ((w + scale_w - 1) // scale_w)

    if level == 0:
        return pw // (1 << (state["dwt_depth_ho"] + state["dwt_depth"]))
    elif level <= state["dwt_depth_ho"]:
        return pw // (1 << (state["dwt_depth_ho"] + state["dwt_depth"] - level + 1))
    elif level > state["dwt_depth_ho"]:
        return pw // (1 << (state["dwt_depth_ho"] + state["dwt_depth"] - level + 1))


@ref_pseudocode(deviation="inferred_implementation")
def subband_height(state, level, comp):
    """(13.2.3)"""
    if comp == "Y":
        h = state["luma_height"]
    elif comp == "C1" or comp == "C2":
        h = state["color_diff_height"]

    # Round up (pad) the picture height to the nearest multiple of the scale
    # height
    scale_h = 1 << state["dwt_depth"]
    ph = scale_h * ((h + scale_h - 1) // scale_h)

    if level == 0:
        return ph // (1 << state["dwt_depth"])
    elif level <= state["dwt_depth_ho"]:
        return ph // (1 << state["dwt_depth"])
    elif level > state["dwt_depth_ho"]:
        return ph // (1 << (state["dwt_depth_ho"] + state["dwt_depth"] - level + 1))


@ref_pseudocode
def slice_bytes(state, sx, sy):
    """(13.5.3.2) Compute the number of bytes in a low-delay picture slice."""
    slice_number = (sy * state["slices_x"]) + sx
    bytes = ((slice_number + 1) * state["slice_bytes_numerator"]) // state[
        "slice_bytes_denominator"
    ]
    bytes -= (slice_number * state["slice_bytes_numerator"]) // state[
        "slice_bytes_denominator"
    ]
    return bytes


@ref_pseudocode
def slice_left(state, sx, c, level):
    """(13.5.6.2) Get the x coordinate of the LHS of the given slice."""
    return (subband_width(state, level, c) * sx) // state["slices_x"]


@ref_pseudocode
def slice_right(state, sx, c, level):
    """(13.5.6.2) Get the x coordinate of the RHS of the given slice."""
    return (subband_width(state, level, c) * (sx + 1)) // state["slices_x"]


@ref_pseudocode
def slice_top(state, sy, c, level):
    """(13.5.6.2) Get the y coordinate of the top of the given slice."""
    return (subband_height(state, level, c) * sy) // state["slices_y"]


@ref_pseudocode
def slice_bottom(state, sy, c, level):
    """(13.5.6.2) Get the y coordinate of the bottom of the given slice."""
    return (subband_height(state, level, c) * (sy + 1)) // state["slices_y"]


def slices_have_same_dimensions(state):
    """
    Utility function, not part of the standard. Tests whether all picture
    slices will have the same dimensions.
    """
    dc_luma_width = subband_width(state, 0, "Y")
    dc_luma_height = subband_height(state, 0, "Y")
    dc_color_diff_width = subband_width(state, 0, "C1")
    dc_color_diff_height = subband_height(state, 0, "C1")
    return (
        dc_luma_width % state["slices_x"] == 0
        and dc_luma_height % state["slices_y"] == 0
        and dc_color_diff_width % state["slices_x"] == 0
        and dc_color_diff_height % state["slices_y"] == 0
    )



"""
The :py:mod:`vc2_conformance.decoder.picture_syntax` module contains pseudocode
functions from (12) Picture syntax.
"""

from vc2_conformance.pseudocode.metadata import ref_pseudocode

from vc2_conformance.constraint_table import allowed_values_for

from vc2_conformance.pseudocode.slice_sizes import slices_have_same_dimensions

from vc2_conformance.pseudocode.parse_code_functions import is_ld, is_hq

from vc2_conformance.decoder.io import (
    tell,
    byte_align,
    read_uint_lit,
    read_uint,
    read_bool,
)

from vc2_data_tables import (
    WaveletFilters,
    QUANTISATION_MATRICES,
)

from vc2_conformance.level_constraints import LEVEL_CONSTRAINTS

from vc2_conformance.version_constraints import (
    wavelet_transform_version_implication,
)

from vc2_conformance.decoder.assertions import (
    assert_in,
    assert_in_enum,
    assert_level_constraint,
    assert_picture_number_incremented_as_expected,
    log_version_lower_bound,
)

from vc2_conformance.decoder.exceptions import (
    ZeroSlicesInCodedPicture,
    SliceBytesHasZeroDenominator,
    SliceBytesIsLessThanOne,
    NoQuantisationMatrixAvailable,
    BadWaveletIndex,
    BadHOWaveletIndex,
    QuantisationMatrixValueNotAllowedInLevel,
    SliceSizeScalerIsZero,
)

from vc2_conformance.decoder.transform_data_syntax import transform_data


__all__ = [
    "picture_parse",
    "picture_header",
    "wavelet_transform",
    "transform_parameters",
    "extended_transform_parameters",
    "slice_parameters",
    "quant_matrix",
]


@ref_pseudocode
def picture_parse(state):
    """(12.1)"""
    byte_align(state)
    picture_header(state)
    byte_align(state)
    wavelet_transform(state)


@ref_pseudocode
def picture_header(state):
    """(12.2)"""
    state["picture_number"] = read_uint_lit(state, 4)

    # (12.2) Checks that the picture incremented by one, wraps correctly and
    # fields have even-numbered picture numbers first.
    assert_picture_number_incremented_as_expected(state, tell(state))  ## Not in spec


@ref_pseudocode
def wavelet_transform(state):
    """(12.3)"""
    transform_parameters(state)
    byte_align(state)
    transform_data(state)


@ref_pseudocode
def transform_parameters(state):
    """(12.4.1)"""
    state["wavelet_index"] = read_uint(state)

    # (12.4.1) Check wavelet type is supported
    ## Begin not in spec
    assert_in_enum(state["wavelet_index"], WaveletFilters, BadWaveletIndex)
    ## End not in spec

    # (C.3) Check level allows this wavelet type
    ## Begin not in spec
    assert_level_constraint(state, "wavelet_index", state["wavelet_index"])
    ## End not in spec

    state["dwt_depth"] = read_uint(state)

    # (C.3) Check level allows this wavelet depth
    assert_level_constraint(state, "dwt_depth", state["dwt_depth"])  ## Not in spec

    state["wavelet_index_ho"] = state["wavelet_index"]
    state["dwt_depth_ho"] = 0

    if state["major_version"] >= 3:
        extended_transform_parameters(state)

    slice_parameters(state)

    quant_matrix(state)


@ref_pseudocode
def extended_transform_parameters(state):
    """(12.4.4.1)"""
    asym_transform_index_flag = read_bool(state)
    # (C.3) Check level allows an asymmetric transform types
    ## Begin not in spec
    assert_level_constraint(
        state, "asym_transform_index_flag", asym_transform_index_flag
    )
    ## End not in spec

    if asym_transform_index_flag:
        state["wavelet_index_ho"] = read_uint(state)

        # (12.4.4.2) Check wavelet type is supported
        ## Begin not in spec
        assert_in_enum(state["wavelet_index_ho"], WaveletFilters, BadHOWaveletIndex)
        ## End not in spec

        # (C.3) Check level allows this wavelet type
        ## Begin not in spec
        assert_level_constraint(state, "wavelet_index_ho", state["wavelet_index_ho"])
        ## End not in spec

    asym_transform_flag = read_bool(state)
    # (C.3) Check level allows an asymmetric transform depths
    ## Begin not in spec
    assert_level_constraint(state, "asym_transform_flag", asym_transform_flag)
    ## End not in spec

    if asym_transform_flag:
        state["dwt_depth_ho"] = read_uint(state)

        # (C.3) Check level allows this wavelet depth
        ## Begin not in spec
        assert_level_constraint(state, "dwt_depth_ho", state["dwt_depth_ho"])
        ## End not in spec

    # (11.2.2) Log increased minimum version requirements if an asymmetric
    # transform was used.
    ## Begin not in spec
    minimum_required_version = wavelet_transform_version_implication(
        state["wavelet_index"],
        state["wavelet_index_ho"],
        state["dwt_depth_ho"],
    )
    log_version_lower_bound(state, minimum_required_version)
    ## End not in spec


@ref_pseudocode
def slice_parameters(state):
    """(12.4.5.2)"""
    state["slices_x"] = read_uint(state)
    # (C.3) Check level allows this number of horizontal slices
    assert_level_constraint(state, "slices_x", state["slices_x"])  ## Not in spec

    state["slices_y"] = read_uint(state)
    # (C.3) Check level allows this number of vertical slices
    assert_level_constraint(state, "slices_y", state["slices_y"])  ## Not in spec

    # Errata: Spec doesn't prevent zero-slice pictures
    #
    # (12.4.5.2) Must have at least once slice
    ## Begin not in spec
    if state["slices_x"] == 0 or state["slices_y"] == 0:
        raise ZeroSlicesInCodedPicture(state["slices_x"], state["slices_y"])
    ## End not in spec

    # (C.3) Check if this level expects slice counts which result in every
    # slice having the same number of transform components
    ## Begin not in spec
    assert_level_constraint(
        state, "slices_have_same_dimensions", slices_have_same_dimensions(state)
    )
    ## End not in spec

    if is_ld(state):
        state["slice_bytes_numerator"] = read_uint(state)
        # (C.3) Check level allows the specified numerator
        ## Begin not in spec
        assert_level_constraint(
            state, "slice_bytes_numerator", state["slice_bytes_numerator"]
        )
        ## End not in spec

        state["slice_bytes_denominator"] = read_uint(state)
        # (C.3) Check level allows the specified denominator
        ## Begin not in spec
        assert_level_constraint(
            state, "slice_bytes_denominator", state["slice_bytes_denominator"]
        )
        ## End not in spec

        # Errata: spec doesn't disallow a slice_bytes_denominator of zero
        #
        # (12.4.5.2) Must have non-zero slice_bytes denominator (i.e. no divide
        # by zero)
        ## Begin not in spec
        if state["slice_bytes_denominator"] == 0:
            raise SliceBytesHasZeroDenominator(state["slice_bytes_numerator"])
        ## End not in spec

        # Errata: spec doesn't disallow slice_bytes_* values which evaluate to
        # less than one byte.
        #
        # (12.4.5.2) slice_bytes_numerator/slice_bytes_denominator must be
        # greater than or equal to one byte to avoid zero-byte slices in the
        # bitstream.
        ## Begin not in spec
        if state["slice_bytes_numerator"] < state["slice_bytes_denominator"]:
            raise SliceBytesIsLessThanOne(
                state["slice_bytes_numerator"],
                state["slice_bytes_denominator"],
            )
        ## End not in spec

    if is_hq(state):
        state["slice_prefix_bytes"] = read_uint(state)
        # (C.3) Check level allows the specified prefix byte count
        ## Begin not in spec
        assert_level_constraint(
            state, "slice_prefix_bytes", state["slice_prefix_bytes"]
        )
        ## End not in spec

        state["slice_size_scaler"] = read_uint(state)

        # Errata: not enforced in spec
        #
        # (12.4.5.2) slice size scaler should not be zero
        ## Begin not in spec
        if state["slice_size_scaler"] == 0:
            raise SliceSizeScalerIsZero()
        ## End not in spec

        # (C.3) Check level allows the specified slice size scaler
        ## Begin not in spec
        assert_level_constraint(state, "slice_size_scaler", state["slice_size_scaler"])
        ## End not in spec


@ref_pseudocode
def quant_matrix(state):
    """(12.4.5.3)"""
    custom_quant_matrix = read_bool(state)
    # (C.3) Check level allows use of custom quantisation matrices
    ## Begin not in spec
    assert_level_constraint(state, "custom_quant_matrix", custom_quant_matrix)
    ## End not in spec

    if custom_quant_matrix:
        # (C.3) Check that quantisation matrix values are in the level-defined
        # range
        ## Begin not in spec
        allowed_values = allowed_values_for(
            LEVEL_CONSTRAINTS,
            "quant_matrix_values",
            state["_level_constrained_values"],
        )

        def check(quant_matrix_value):
            assert_in(
                quant_matrix_value,
                allowed_values,
                QuantisationMatrixValueNotAllowedInLevel,
                state["_level_constrained_values"],
            )

        ## End not in spec

        # NB: For historical reasons, we use a dict not an array in this
        # implementation.
        ### state["quant_matrix"] = new_array(state["dwt_depth_ho"] + state["dwt_depth"] + 1)
        state["quant_matrix"] = {}  ## Not in spec
        if state["dwt_depth_ho"] == 0:
            state["quant_matrix"][0] = {}
            state["quant_matrix"][0]["LL"] = read_uint(state)
            check(state["quant_matrix"][0]["LL"])  ## Not in spec
        else:
            state["quant_matrix"][0] = {}
            state["quant_matrix"][0]["L"] = read_uint(state)
            check(state["quant_matrix"][0]["L"])  ## Not in spec
            for level in range(1, state["dwt_depth_ho"] + 1):
                state["quant_matrix"][level] = {}
                state["quant_matrix"][level]["H"] = read_uint(state)
                check(state["quant_matrix"][level]["H"])  ## Not in spec
        for level in range(
            state["dwt_depth_ho"] + 1, state["dwt_depth_ho"] + state["dwt_depth"] + 1
        ):
            state["quant_matrix"][level] = {}
            state["quant_matrix"][level]["HL"] = read_uint(state)
            check(state["quant_matrix"][level]["HL"])  ## Not in spec
            state["quant_matrix"][level]["LH"] = read_uint(state)
            check(state["quant_matrix"][level]["LH"])  ## Not in spec
            state["quant_matrix"][level]["HH"] = read_uint(state)
            check(state["quant_matrix"][level]["HH"])  ## Not in spec
    else:
        # (12.4.5.3) If no custom quantisation matrix has been specified, a
        # default table must be available
        ## Begin not in spec
        configuration = (
            state["wavelet_index"],
            state["wavelet_index_ho"],
            state["dwt_depth"],
            state["dwt_depth_ho"],
        )
        if configuration not in QUANTISATION_MATRICES:
            raise NoQuantisationMatrixAvailable(*configuration)
        ## End not in spec

        set_quant_matrix(state)


@ref_pseudocode(deviation="inferred_implementation")
def set_quant_matrix(state):
    """(12.4.5.3)"""
    state["quant_matrix"] = QUANTISATION_MATRICES[
        (
            state["wavelet_index"],
            state["wavelet_index_ho"],
            state["dwt_depth"],
            state["dwt_depth_ho"],
        )
    ]
